@@ -1,13 +1,24 @@
-/* Copyright (C) 1989, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevdjet.c */
@@ -19,7 +30,7 @@
  * Thanks for various improvements to:
  *	Jim Mayer (mayer@wrc.xerox.com)
  *	Jan-Mark Wams (jms@cs.vu.nl)
- *	Frans van Hoesel (hoesel@rugr86.rug.nl)
+ *	Frans van Hoesel (hoesel@chem.rug.nl)
  *	George Cameron (g.cameron@biomed.abdn.ac.uk)
  *	Nick Duffek (nsd@bbc.com)
  * Thanks for the LJIIID duplex capability to:
@@ -36,39 +47,8 @@
  * If the preprocessor symbol A4 is defined, the default paper size is
  * the European A4 size; otherwise it is the U.S. letter size (8.5"x11").
  *
- * You may find the following test page useful in determining the exact
- * margin settings on your printer.  It prints four big arrows which
- * point exactly to the for corners of an A4 sized paper. Of course the
- * arrows cannot appear in full on the paper, and they are truncated by
- * the margins. The margins measured on the testpage must match those
- * in gdevdjet.c.  So the testpage indicates two facts: 1) the page is
- * not printed in the right position 2) the page is truncated too much
- * because the margins are wrong. Setting wrong margins in gdevdjet.c
- * will also move the page, so both facts should be matched with the
- * real world.
-
-%!
-	newpath 
-	0 0 moveto 144 72 lineto 72 144 lineto
-	closepath fill stroke 0 0 moveto 144 144 lineto stroke
-
-	595.27 841.88 moveto 451.27 769.88 lineto 523.27 697.88 lineto
-	closepath fill stroke 595.27 841.88 moveto 451.27 697.88 lineto stroke
-
-	0 841.88 moveto 144 769.88 lineto 72 697.88 lineto
-	closepath fill stroke 0 841.88 moveto 144 697.88 lineto stroke
-
-	595.27 0 moveto 451.27 72 lineto 523.27 144 lineto
-	closepath fill stroke 595.27 0 moveto 451.27 144 lineto stroke
-
-	/Helvetica findfont
-	14 scalefont setfont
-	100 600 moveto
-	(This is an A4 testpage. The arrows should point exactly to the) show
-	100 580 moveto
-	(corners and the margins should match those given in gdev*.c) show
-	showpage
-
+ * To determine the proper "margin" settings for your printer, see the
+ * file align.ps.
  */
 
 /* Define the default, maximum resolutions. */
@@ -108,13 +88,17 @@
  * inexact paperlength which is set to 117 10ths.
  * Somebody should check for letter sized paper. I left it at 0.07".
  */
-#define DESKJET_MARGINS_LETTER  0.25, 0.50, 0.25, 0.07
+#define DESKJET_MARGINS_LETTER  0.2, 0.45, 0.3, 0.05
 #define DESKJET_MARGINS_A4	0.125, 0.5, 0.143, 0.09
 /* Similar margins for the LaserJet, */
-/* from Eddy Andrews eeandrew@pyr.swan.ac.uk; */
+/* from Eddy Andrews eeandrew@pyr.swan.ac.uk: */
 #define LASERJET_MARGINS_A4	0.25, 0.20, 0.25, 0.00
-/* from Dale Atems atems@igor.physics.wayne.edu. */
-#define LASERJET_MARGINS_LETTER	0.275, -0.06, 0.425, 0.26
+/* the following from Bob Atkins bob@digilink.net are WRONG: */
+/*#define LASERJET_MARGINS_A4	0.25, 0.0, 0.25, 0.0*/
+/* from Dale Atems atems@igor.physics.wayne.edu: */
+/*#define LASERJET_MARGINS_LETTER	0.275, -0.06, 0.425, 0.26*/
+/* ... superseded by the following from Bob Atkins bob@digilink.net: */
+#define LASERJET_MARGINS_LETTER	0.25, 0.0, 0.25, 0.0
 
 /* The number of blank lines that make it worthwhile to reposition */
 /* the cursor. */
@@ -137,7 +121,7 @@
 
 /*
  * The notion that there is such a thing as a "PCL printer" is a fiction:
- * no two "PCL" printers, even at the same PCL level, have compatible
+ * no two "PCL" printers, even at the same PCL level, have identical
  * command sets.  The command strings below were established by hearsay
  * and by trial and error.  (The H-P documentation isn't fully accurate
  * either; for example, it doesn't reveal that the DeskJet printers
@@ -253,19 +237,20 @@ private int
 hpjet_open(gx_device *pdev)
 {	/* Change the margins if necessary. */
 	const float _ds *m = 0;
-	if ( ppdev->print_page == djet_print_page ||
-	     ppdev->print_page == djet500_print_page
+
+	if ( ppdev->printer_procs.print_page == djet_print_page ||
+	     ppdev->printer_procs.print_page == djet500_print_page
 	   )
 	{	static const float m_a4[4] = { DESKJET_MARGINS_A4 };
 		static const float m_letter[4] = { DESKJET_MARGINS_LETTER };
 		m = (gdev_pcl_paper_size(pdev) == PAPER_SIZE_A4 ? m_a4 :
 		     m_letter);
 	}
-	else if ( ppdev->print_page == oce9050_print_page )
-	{	static const float m_none[4] = { 0.00, 0.00, 0.00, 0.00 };
-		m = m_none;
-	}
-	else if ( ppdev->print_page != lp2563_print_page )	/* LaserJet */
+	else if ( ppdev->printer_procs.print_page == oce9050_print_page ||
+		  ppdev->printer_procs.print_page == lp2563_print_page
+		)
+	  ;
+	else		/* LaserJet */
 	{	static const float m_a4[4] = { LASERJET_MARGINS_A4 };
 		static const float m_letter[4] = { LASERJET_MARGINS_LETTER };
 		m = (gdev_pcl_paper_size(pdev) == PAPER_SIZE_A4 ? m_a4 :
@@ -273,18 +258,20 @@ hpjet_open(gx_device *pdev)
 	}
 	if ( m != 0 )
 	  gx_device_set_margins(pdev, m, true);
+	/* If this is a LJIIID, enable Duplex. */
+	if ( ppdev->printer_procs.print_page == ljet3d_print_page )
+	  ppdev->Duplex = true, ppdev->Duplex_set = 0;
 	return gdev_prn_open(pdev);
 }
 
 /* hpjet_close is only here to eject odd numbered pages in duplex mode. */
 private int
 hpjet_close(gx_device *pdev)
-{	if ( pdev->Duplex_set >= 0 && pdev->Duplex )
+{	if ( ppdev->Duplex_set >= 0 && ppdev->Duplex )
 	  {	gdev_prn_open_printer(pdev, 1);
 		fputs("\033&l0H", ppdev->file) ;
-		return gdev_prn_close_printer(pdev);
 	  }
-	return 0;
+	return gdev_prn_close(pdev);
 }
 
 #undef ppdev
@@ -295,44 +282,44 @@ hpjet_close(gx_device *pdev)
 private int
 djet_print_page(gx_device_printer *pdev, FILE *prn_stream)
 {	return hpjet_print_page(pdev, prn_stream, DJ, 300, mode_2,
-		"\033&k1W\033*p0x0Y\033*b2M");
+		"\033&k1W\033*b2M");
 }
 /* The DeskJet500 can compress (modes 2&3) */
 private int
 djet500_print_page(gx_device_printer *pdev, FILE *prn_stream)
 {	return hpjet_print_page(pdev, prn_stream, DJ500, 300, mode_3,
-		"\033&k1W\033*p0x0Y");
+		"\033&k1W");
 }
 /* The LaserJet series II can't compress */
 private int
 ljet_print_page(gx_device_printer *pdev, FILE *prn_stream)
 {	return hpjet_print_page(pdev, prn_stream, LJ, 300, mode_0,
-		"\033*p0x0Y\033*b0M");
+		"\033*b0M");
 }
 /* The LaserJet Plus can't compress */
 private int
 ljetplus_print_page(gx_device_printer *pdev, FILE *prn_stream)
 {	return hpjet_print_page(pdev, prn_stream, LJplus, 300, mode_0,
-		"\033*p0x0Y\033*b0M");
+		"\033*b0M");
 }
 /* LaserJet series IIp & IId compress (mode 2) */
 /* but don't support *p+ or *b vertical spacing. */
 private int
 ljet2p_print_page(gx_device_printer *pdev, FILE *prn_stream)
 {	return hpjet_print_page(pdev, prn_stream, LJ2p, 300, mode_2p,
-		"\033*r0F\033*p0x0Y\033*b2M");
+		"\033*r0F\033*b2M");
 }
 /* All LaserJet series IIIs (III,IIId,IIIp,IIIsi) compress (modes 2&3) */
 private int
 ljet3_print_page(gx_device_printer *pdev, FILE *prn_stream)
 {	return hpjet_print_page(pdev, prn_stream, LJ3, 300, mode_3,
-		"\033*r0F\033*p0x0Y");
+		"\033*r0F");
 }
 /* LaserJet IIId is same as LaserJet III, except for duplex */
 private int
 ljet3d_print_page(gx_device_printer *pdev, FILE *prn_stream)
 {	return hpjet_print_page(pdev, prn_stream, LJ3D, 300, mode_3,
-		"\033*r0F\033*p0x0Y");
+		"\033*r0F");
 }
 /* LaserJet 4 series compresses, and it needs a special sequence to */
 /* allow it to specify coordinates at 600 dpi. */
@@ -340,7 +327,7 @@ private int
 ljet4_print_page(gx_device_printer *pdev, FILE *prn_stream)
 {	int dots_per_inch = (int)pdev->y_pixels_per_inch;
 	char real_init[50];
-	sprintf(real_init, "\033*r0F\033*p0x0Y\033&u%dD", dots_per_inch);
+	sprintf(real_init, "\033*r0F\033&u%dD", dots_per_inch);
 	return hpjet_print_page(pdev, prn_stream, LJ4, dots_per_inch, mode_3,
 		real_init);
 }
@@ -349,7 +336,7 @@ ljet4_print_page(gx_device_printer *pdev, FILE *prn_stream)
 private int
 lp2563_print_page(gx_device_printer *pdev, FILE *prn_stream)
 {	return hpjet_print_page(pdev, prn_stream, LP2563B, 300, mode_0ns,
-		"\033*p0x0Y\033*b0M");
+		"\033*b0M");
 }
 /* The Oce line printer has TIFF compression */
 /* and doesn't support *p+ or *b vertical spacing. */
@@ -364,7 +351,7 @@ oce9050_print_page(gx_device_printer *pdev, FILE *prn_stream)
 	fputs("\033%1A", prn_stream);	/* Enter PCL mode */
 
 	code = hpjet_print_page(pdev, prn_stream, OCE9050, 400, mode_3ns,
-				"\033*p0x0Y\033*b3M");
+				"\033*b3M");
 
 	/* Return to HPGL/2 mode. */
 	fputs("\033%1B", prn_stream);	/* Enter HPGL/2 mode */
@@ -445,8 +432,16 @@ hpjet_print_page(gx_device_printer *pdev, FILE *prn_stream, int ptype,
 		}
 	}
 
-	fputs("\033*rB", prn_stream);		/* end raster graphics */
-	fprintf(prn_stream, "\033*t%dR", x_dpi);	/* set resolution */
+		/* End raster graphics, position cursor at top. */
+	fputs("\033*rB\033*p0x0Y", prn_stream);
+
+		/* The DeskJet and DeskJet Plus reset everything upon */
+		/* receiving \033*rB, so we must reinitialize graphics mode. */
+	if ( ptype == DJ ) 
+          fputs(init_string, prn_stream);
+
+		/* Set resolution. */
+	fprintf(prn_stream, "\033*t%dR", x_dpi);
 
 	/* Send each scan line in turn */
 	   {	int lnum;
