@@ -1,13 +1,24 @@
-/* Copyright (C) 1992, 1993, 1994 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1992, 1993, 1994, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevpm.c */
@@ -76,6 +87,8 @@ typedef struct gx_device_pm_s gx_device_pm;
 
 #define gx_device_pm_common\
 	int BitsPerPixel;\
+	int alpha_text;\
+	int alpha_graphics;\
 	int UpdateInterval;\
 	char GSVIEW[pm_gsview_sizeof];\
 	BOOL dll;\
@@ -114,8 +127,10 @@ private dev_proc_map_color_rgb(pm_map_color_rgb);
 private dev_proc_fill_rectangle(pm_fill_rectangle);
 private dev_proc_copy_mono(pm_copy_mono);
 private dev_proc_copy_color(pm_copy_color);
+private dev_proc_get_bits(pm_get_bits);
 private dev_proc_get_params(pm_get_params);
 private dev_proc_put_params(pm_put_params);
+private dev_proc_get_alpha_bits(pm_get_alpha_bits);
 
 private gx_device_procs pm_procs = {
 	pm_open,
@@ -126,34 +141,40 @@ private gx_device_procs pm_procs = {
 	pm_map_rgb_color,
 	pm_map_color_rgb,
 	pm_fill_rectangle,
-	gx_default_tile_rectangle,
+	NULL,			/* tile rectangle */
 	pm_copy_mono,
 	pm_copy_color,
-	gx_default_draw_line,
-	gx_default_get_bits,
+	NULL,			/* draw line */
+	pm_get_bits,
 	pm_get_params,
 	pm_put_params,
-	gx_default_map_cmyk_color,
-	gx_default_get_xfont_procs
+	NULL,			/* map_cmyk_color */
+	gx_default_get_xfont_procs,
+	NULL,			/* get_xfont_device */
+	NULL,			/* map_rgb_alpha_color */
+	gx_page_device_get_page_device,
+	pm_get_alpha_bits
 };
 #ifdef __DLL__
-gx_device_pm gs_os2dll_device = {
+gx_device_pm far_data gs_os2dll_device = {
 	std_device_std_body(gx_device_pm, &pm_procs, "os2dll",
 	  INITIAL_WIDTH, INITIAL_HEIGHT,
 	  INITIAL_RESOLUTION, INITIAL_RESOLUTION),
 	 { 0 },		/* std_procs */
 	8,		/* BitsPerPixel */
+	1, 1,		/* alpha */
 	5000,		/* UpdateInterval */
 	"\0",		/* GSVIEW */
 	1		/* is DLL device */
 };
 #endif
-gx_device_pm gs_os2pm_device = {
+gx_device_pm far_data gs_os2pm_device = {
 	std_device_std_body(gx_device_pm, &pm_procs, "os2pm",
 	  INITIAL_WIDTH, INITIAL_HEIGHT,
 	  INITIAL_RESOLUTION, INITIAL_RESOLUTION),
 	 { 0 },		/* std_procs */
 	8,		/* BitsPerPixel */
+	1, 1,		/* alpha */
 	5000,		/* UpdateInterval */
 	"\0",		/* GSVIEW */
 	0		/* is not DLL device */
@@ -193,24 +214,42 @@ pm_open(gx_device *dev)
 		fprintf(stderr,"\npm_open: Couldn't get pid\n");
 		return gs_error_limitcheck;
 	}
-	pmdev->gspid = pppib->pib_ulppid; /* use parent (CMD.EXE) pid */
+#ifdef __DLL__
+	if (pppib->pib_ultype == 3)	     /* if caller is PM app */
+	    pmdev->gspid = pppib->pib_ulpid; /* use caller pid */
+	else
+#endif
+	    pmdev->gspid = pppib->pib_ulppid; /* use parent (CMD.EXE) pid */
 	sprintf(id, ID_NAME, pmdev->gspid, (ULONG)dev);
 
 	/* Allocate, but don't commit, enough memory for the largest */
 	/* possible bitmap (13Mbytes = A3 x 150dpi x 24bits) */
-	/* Shared memory is common to all processes so we don't want to allocate too much */
-	sprintf(name, SHARED_NAME, *pmdev->GSVIEW ? pmdev->GSVIEW : id);
-	if (DosAllocSharedMem((PPVOID)&pmdev->bitmap, name,
+#ifdef __DLL__
+	if (pmdev->dll) {
+	    /* We don't need to use shared memory for the DLL */
+	    if (DosAllocMem((PPVOID)&pmdev->bitmap, 
+		13*1024*1024, PAG_READ | PAG_WRITE)) {
+		fprintf(stderr,"pm_open: failed allocating BMP memory\n");
+		return gs_error_limitcheck;
+	    }
+	}
+	else 
+#endif
+	{
+	    /* Shared memory is common to all processes so we don't want to allocate too much */
+	    sprintf(name, SHARED_NAME, *pmdev->GSVIEW ? pmdev->GSVIEW : id);
+	    if (DosAllocSharedMem((PPVOID)&pmdev->bitmap, name,
 		13*1024*1024, PAG_READ | PAG_WRITE)) {
 		fprintf(stderr,"pm_open: failed allocating shared BMP memory %s\n", name);
 		return gs_error_limitcheck;
+	    }
 	}
 	
 	/* commit one page so there is enough storage for a */
 	/* bitmap header and palette */
 	if (DosSetMem(pmdev->bitmap, MIN_COMMIT, PAG_COMMIT | PAG_DEFAULT)) {
 		DosFreeMem(pmdev->bitmap);
-		fprintf(stderr,"pm_open: failed committing shared BMP memory\n");
+		fprintf(stderr,"pm_open: failed committing BMP memory\n");
 		return gs_error_limitcheck;
 	}
 	pmdev->committed = MIN_COMMIT;
@@ -479,7 +518,9 @@ pm_close(gx_device *dev)
 APIRET rc;
 #ifdef __DLL__
     if (pmdev->dll) {
+	DosRequestMutexSem(pmdev->bmp_mutex, 60000);
         (*pgsdll_callback)(GSDLL_DEVICE, (unsigned char *)dev, 0);
+	DosReleaseMutexSem(pmdev->bmp_mutex);
     }
     else
 #endif
@@ -510,6 +551,8 @@ APIRET rc;
 	    /* wait for termination queue, queue is then closed by session manager */
 	    DosReadQueue(pmdev->term_queue, &Request, &DataLength, 
 		&DataAddress, 0, DCWW_WAIT, &ElemPriority, (HEV)NULL);
+	    /* queue needs to be closed by us */
+	    DosCloseQueue(pmdev->term_queue);
 	}
     }
     /* release memory */
@@ -542,17 +585,11 @@ gx_color_index
 pm_map_rgb_color(gx_device *dev, gx_color_value r, gx_color_value g,
   gx_color_value b)
 {
-	if (!dev->is_open)
-	    return 0;
-	switch(pmdev->BitsPerPixel) {
+	switch(dev->color_info.depth) {
 	  case 24:
 		return	((b >> (gx_color_value_bits - 8)) << 16) +
 			((g >> (gx_color_value_bits - 8)) << 8) +
 			((r >> (gx_color_value_bits - 8)));
-	  case 16:
-		return	(pm_color_value(b) << 16) +
-			(pm_color_value(g) << 8) +
-			(pm_color_value(r));
 	  case 8: {
 		int i;
 		RGB2 *prgb;
@@ -605,24 +642,24 @@ pm_map_rgb_color(gx_device *dev, gx_color_value r, gx_color_value g,
 int
 pm_map_color_rgb(gx_device *dev, gx_color_index color,
   gx_color_value prgb[3])
-{
-gx_color_value one;
-	if (!dev->is_open)
-	    return 0;
-	switch(pmdev->BitsPerPixel) {
+{	gx_color_value one;
+	switch(dev->color_info.depth) {
 	  case 24:
-	  case 16:
-		/* untested */
 		one = (gx_color_value) (gx_max_color_value / 255);
 		prgb[0] = ((color)     & 255) * one;
 		prgb[1] = ((color>>8)  & 255) * one;
 		prgb[2] = ((color>>16) & 255) * one;
 		break;
 	  case 8:
+		if (!dev->is_open)
+		    return -1;
+		{
+		RGB2 *argb = (RGB2 *)((PBYTE)pmdev->bmi + pmdev->bmi->cbFix);
 		one = (gx_color_value) (gx_max_color_value / 255);
-		prgb[0] = pmdev->bmi->argbColor[(int)color].bRed * one;
-		prgb[1] = pmdev->bmi->argbColor[(int)color].bGreen * one;
-		prgb[2] = pmdev->bmi->argbColor[(int)color].bBlue * one;
+		prgb[0] = argb[(int)color].bRed * one;
+		prgb[1] = argb[(int)color].bGreen * one;
+		prgb[2] = argb[(int)color].bBlue * one;
+		}
 		break;
 	  case 4:
 		if (color == 8)	/* VGA light gray */
@@ -677,6 +714,11 @@ pm_copy_color(gx_device *dev,
 	return 0;
 }
 
+int
+pm_get_bits(gx_device *dev, int y, byte *str, byte **actual_data)
+{
+	return pmmproc(get_bits)(pmmdev, y, str, actual_data);
+}
 /* Get PM parameters */
 int
 pm_get_params(gx_device *dev, gs_param_list *plist)
@@ -690,6 +732,33 @@ pm_get_params(gx_device *dev, gs_param_list *plist)
 	return code;
 }
 
+/* Put parameters. */
+private int
+pm_put_alpha_param(gs_param_list *plist, gs_param_name param_name, int *pa,
+  bool alpha_ok)
+{	int code = param_read_int(plist, param_name, pa);
+	switch ( code )
+	{
+	case 0:
+		switch ( *pa )
+		  {
+		  case 1:
+			return 0;
+		  case 2: case 4:
+			if ( alpha_ok )
+			  return 0;
+		  default:
+			code = gs_error_rangecheck;
+		  }
+	default:
+		param_signal_error(plist, param_name, code);
+	case 1:
+		;
+	}
+	return code;
+}
+
+
 /* Set PM parameters -- size and resolution. */
 /* We implement this ourselves so that we can do it without */
 /* closing and opening the device. */
@@ -705,6 +774,8 @@ pm_put_params(gx_device *dev, gs_param_list *plist)
 	int bpp = old_bpp;
 	int uii = pmdev->UpdateInterval;
 	gs_param_string gsvs;
+	int atext = pmdev->alpha_text, agraphics = pmdev->alpha_graphics;
+	bool alpha_ok;
 
 	/* Handle extra parameters */
 	switch ( code = param_read_string(plist, "GSVIEW", &gsvs) )
@@ -766,24 +837,24 @@ bppe:		param_signal_error(plist, "BitsPerPixel", ecode);
 		break;
 	}
 
+	alpha_ok = pmdev->color_info.depth >= 8;
+	if ( (code = pm_put_alpha_param(plist, "TextAlphaBits", &pmdev->alpha_text, alpha_ok)) < 0 )
+	  ecode = code;
+	if ( (code = pm_put_alpha_param(plist, "GraphicsAlphaBits", &pmdev->alpha_graphics, alpha_ok)) < 0 )
+	  ecode = code;
+
 	if ( ecode >= 0 )
-	  {
-		static const gx_device_color_info pm_fake_16bit_color = dci_color(16,31,31);
-		gx_device_color_info old_color_info = dev->color_info;
-		/* prevent default routine from objecting to -dBitsPerPixe=16 */
-		/* when underlying bitmap is 24bits/pixel */
-		if (bpp == 16)
-		    dev->color_info = pm_fake_16bit_color;
-		/* Prevent gx_default_put_params from closing the device. */
+	  {	/* Prevent gx_default_put_params from closing the device. */
 		dev->is_open = false;
 		ecode = gx_default_put_params(dev, plist);
 		dev->is_open = is_open;
-		dev->color_info = old_color_info;
 	  }
 	if ( ecode < 0 )
 	  {
 		if ( bpp != old_bpp )
-		  pm_set_bits_per_pixel(pmdev, bpp);
+		  pm_set_bits_per_pixel(pmdev, old_bpp);
+		  pmdev->alpha_text = atext;
+		  pmdev->alpha_graphics = agraphics;
 		return ecode;
 	  }
 
@@ -804,6 +875,8 @@ bppe:		param_signal_error(plist, "BitsPerPixel", ecode);
 			dev->width = width;
 			dev->height = height;
 		  	pm_set_bits_per_pixel(pmdev, old_bpp);
+			pmdev->alpha_text = atext;
+			pmdev->alpha_graphics = agraphics;
 			pm_alloc_bitmap(pmdev, dev);
 			DosReleaseMutexSem(pmdev->bmp_mutex);
 			return ccode;
@@ -847,11 +920,19 @@ bppe:		param_signal_error(plist, "BitsPerPixel", ecode);
 	return 0;
 }
 
+
+/* Get the number of alpha bits. */
+int
+pm_get_alpha_bits(gx_device *dev, graphics_object_type type)
+{	return (type == go_text ? pmdev->alpha_text : pmdev->alpha_graphics);
+}
+
+
 #ifdef __DLL__
 /* ------ DLL routines ------ */
 /* store at pbitmap the address of the bitmap */
 /* device is a pointer to Ghostscript device from GSDLL_DEVICE message */
-unsigned long 
+unsigned long GSDLLAPI
 gsdll_get_bitmap(unsigned char *device, unsigned char **pbitmap)
 {
 gx_device *dev = (gx_device *)device;
@@ -859,11 +940,11 @@ gx_device *dev = (gx_device *)device;
     return 0;
 }
 
-/* Lock the bitmap (so it's size cannot be changed) if flag = TRUE */
-/* or unlock the bitmap if flag = FALSE */
+/* Lock the device (so it's size cannot be changed) if flag = TRUE */
+/* or unlock the device if flag = FALSE */
 /* device is a pointer to Ghostscript device from GSDLL_DEVICE message */
-unsigned long 
-gsdll_lock_bitmap(unsigned char *device, int flag)
+int GSDLLAPI
+gsdll_lock_device(unsigned char *device, int flag)
 {
 gx_device *dev = (gx_device *)device;
 APIRET rc;
@@ -912,7 +993,7 @@ pm_run_gspmdrv(gx_device_pm *pmdev)
 		fprintf(stderr,"pm_run_gspmdrv: Couldn't get module handle, rc = %d\n", rc);
 		return gs_error_limitcheck;
 	}
-	if ( (rc = DosQueryModuleName(pppib->pib_hmte, 256, progname)) != 0 ) {
+	if ( (rc = DosQueryModuleName(pppib->pib_hmte, sizeof(progname)-1, progname)) != 0 ) {
 		fprintf(stderr,"pm_run_gspmdrv: Couldn't get module name, rc = %d\n", rc);
 		return gs_error_limitcheck;
 	}
@@ -1010,6 +1091,18 @@ pm_alloc_bitmap(gx_device_pm *pmdev, gx_device *param_dev)
 		pmdev->committed = needed;
 	}
 	/* Shared memory can't be decommitted */
+#ifdef __DLL__
+	if (needed < pmdev->committed) {
+		/* decommit memory */
+		if (rc = DosSetMem(pmdev->bitmap + needed, 
+			pmdev->committed - needed, 
+			PAG_DECOMMIT)) {
+			fprintf(stderr,"Failed to decommit memory in pm_alloc_bitmap, rc = %d\n",rc);
+			return gs_error_limitcheck;
+		}
+		pmdev->committed = needed;
+	}
+#endif
 
 	/* Nothing can go wrong now.... */
 	base = pmdev->bitmap + pmdev->bmi->cbFix + pm_palette_size(pmdev);
@@ -1093,7 +1186,6 @@ private uint
 pm_set_bits_per_pixel(gx_device_pm *pmdev, int bpp)
 {
 static const gx_device_color_info pm_24bit_color = dci_color(24,255,255);
-static const gx_device_color_info pm_16bit_color = dci_color(24,31,31);
 static const gx_device_color_info pm_8bit_color = dci_color(8,31,4);
 static const gx_device_color_info pm_4bit_color = dci_pc_4bit;
 static const gx_device_color_info pm_2color = dci_black_and_white;
@@ -1102,11 +1194,6 @@ static const gx_device_color_info pm_2color = dci_black_and_white;
 		pmdev->color_info = pm_24bit_color;
 		pmdev->nColors = (1<<24);
 		break;
-	    case 16:
-		/* bitmap is stored as 24 bits/pixel */
-		pmdev->color_info = pm_16bit_color;
-		pmdev->nColors = 32768;
-		break;
 	    case 8:
 		/* use 64 static colors and 166 dynamic colors from 8 planes */
 		pmdev->color_info = pm_8bit_color;
