@@ -1,13 +1,24 @@
-/* Copyright (C) 1991, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1991, 1995, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevsvga.c */
@@ -41,7 +52,7 @@ private dc_entry dynamic_colors[dc_hash_size + 1];
 	svga_map_rgb_color, svga_map_color_rgb,\
 	svga_fill_rectangle, NULL /*tile_rectangle*/,\
 	svga_copy_mono, svga_copy_color, NULL /*draw_line*/,\
-	svga_get_bits, svga_get_params, svga_put_params,\
+	svga_get_bits, NULL /*get_params*/, svga_put_params,\
 	NULL /*map_cmyk_color*/, NULL /*get_xfont_procs*/,\
 	NULL /*get_xfont_device*/, NULL /*map_rgb_alpha_color*/,\
 	NULL /*get_page_device*/, svga_get_alpha_bits, svga_copy_alpha\
@@ -416,16 +427,6 @@ svga_copy_color(gx_device *dev,
 	return 0;
 }
 
-/* Get parameters.  We add TextAlphaBits and GraphicsAlphaBits. */
-int
-svga_get_params(gx_device *dev, gs_param_list *plist)
-{	int code = gx_default_get_params(dev, plist);
-	if ( code < 0 ) return code;
-	code = param_write_int(plist, "TextAlphaBits", &fb_dev->alpha_text);
-	if ( code < 0 ) return code;
-	return param_write_int(plist, "GraphicsAlphaBits", &fb_dev->alpha_graphics);
-}
-
 /* Put parameters. */
 int
 svga_put_params(gx_device *dev, gs_param_list *plist)
@@ -434,7 +435,7 @@ svga_put_params(gx_device *dev, gs_param_list *plist)
 	int atext = fb_dev->alpha_text, agraphics = fb_dev->alpha_graphics;
 	const char _ds *param_name;
 
-	switch ( code = param_read_int(plist, (param_name = "TextAlphaBits"), &atext) )
+	switch ( code = param_read_int(plist, (param_name = "TextAlphaBits"), &fb_dev->alpha_text) )
 	{
 	case 0:
 		if ( atext == 1 || atext == 2 || atext == 4 )
@@ -447,7 +448,7 @@ svga_put_params(gx_device *dev, gs_param_list *plist)
 		;
 	}
 
-	switch ( code = param_read_int(plist, (param_name = "GraphicsAlphaBits"), &agraphics) )
+	switch ( code = param_read_int(plist, (param_name = "GraphicsAlphaBits"), &fb_dev->alpha_graphics) )
 	{
 	case 0:
 		if ( agraphics == 1 || agraphics == 2 || agraphics == 4 )
@@ -460,15 +461,13 @@ svga_put_params(gx_device *dev, gs_param_list *plist)
 		;
 	}
 
-	if ( ecode < 0 )
-	  return ecode;
-	code = gx_default_put_params(dev, plist);
-	if ( code < 0 )
-	  return code;
-
-	fb_dev->alpha_text = atext;
-	fb_dev->alpha_graphics = agraphics;
-	return 0;
+	if ( (code = ecode) < 0 ||
+	     (code = gx_default_put_params(dev, plist)) < 0
+	   )
+	  { fb_dev->alpha_text = atext;
+	    fb_dev->alpha_graphics = agraphics;
+	  }
+	return code;
 }
 
 /* Read scan lines back from the frame buffer. */
@@ -579,7 +578,7 @@ private gx_device_procs vesa_procs = svga_procs(vesa_open);
 int vesa_get_mode(P0());
 void vesa_set_mode(P1(int));
 private void vesa_set_page(P3(gx_device_svga *, int, int));
-gx_device_svga gs_vesa_device =
+gx_device_svga far_data gs_vesa_device =
 	svga_device(vesa_procs, "vesa", vesa_get_mode, vesa_set_mode, vesa_set_page);
 
 /* Define the structures for information returned by the BIOS. */
@@ -769,7 +768,7 @@ private gx_device_procs atiw_procs = svga_procs(atiw_open);
 private int atiw_get_mode(P0());
 private void atiw_set_mode(P1(int));
 private void atiw_set_page(P3(gx_device_svga *, int, int));
-gx_device_svga gs_atiw_device =
+gx_device_svga far_data gs_atiw_device =
 	svga_device(atiw_procs, "atiw", atiw_get_mode, atiw_set_mode, atiw_set_page);
 
 /* Read the device mode */
@@ -827,7 +826,7 @@ private dev_proc_open_device(tvga_open);
 private gx_device_procs tvga_procs = svga_procs(tvga_open);
 /* We can use the atiw_get/set_mode procedures. */
 private void tvga_set_page(P3(gx_device_svga *, int, int));
-gx_device_svga gs_tvga_device =
+gx_device_svga far_data gs_tvga_device =
 	svga_device(tvga_procs, "tvga", atiw_get_mode, atiw_set_mode, tvga_set_page);
 
 /* Initialize the graphics mode. */
@@ -870,7 +869,7 @@ private gx_device_procs tseng_procs =
 private void tseng_set_page(P3(gx_device_svga *, int, int));
 
 /* The 256-color Tseng device */
-gx_device_svga gs_tseng_device =
+gx_device_svga far_data gs_tseng_device =
 	svga_device(tseng_procs, "tseng", atiw_get_mode, atiw_set_mode, tseng_set_page);
 
 /* Initialize the graphics mode. */
