@@ -1,22 +1,31 @@
-/* Copyright (C) 1992, 1993, 1994 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1992, 1995, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevwdib.c */
 /* MS Windows 3.n driver for Ghostscript using a DIB for buffering. */
 #include "gdevmswn.h"
 #include "gxdevmem.h"
-#ifdef __DLL__
 #include "gsdll.h"
-#endif
 
 #ifdef __WIN32__
 #  define USE_SEGMENTS 0
@@ -38,8 +47,9 @@ private dev_proc_close_device(win_dib_close);
 private dev_proc_fill_rectangle(win_dib_fill_rectangle);
 private dev_proc_copy_mono(win_dib_copy_mono);
 private dev_proc_copy_color(win_dib_copy_color);
+private dev_proc_get_bits(win_dib_get_bits);
+private dev_proc_put_params(win_dib_put_params);
 /* Windows-specific procedures */
-private win_proc_copy_to_clipboard(win_dib_copy_to_clipboard);
 private win_proc_repaint(win_dib_repaint);
 private win_proc_alloc_bitmap(win_dib_alloc_bitmap);
 private win_proc_free_bitmap(win_dib_free_bitmap);
@@ -61,6 +71,10 @@ struct gx_device_win_dib_s {
 #endif		/* USE_SEGMENTS */
 
 	HGLOBAL hmdata;
+#ifdef __WIN32__
+	HANDLE hmtx;
+#endif
+	int lock_count;
 	gx_device_memory mdev;
 };
 private gx_device_procs win_dib_procs = {
@@ -76,51 +90,34 @@ private gx_device_procs win_dib_procs = {
 	win_dib_copy_mono,
 	win_dib_copy_color,
 	NULL,			/* draw_line */
-	NULL,			/* get_bits */
+	win_dib_get_bits /* NULL */,			/* get_bits */
 	win_get_params,
-	win_put_params,
+	win_dib_put_params,
 	NULL,			/* map_cmyk_color */
-	win_get_xfont_procs
+	win_get_xfont_procs,
+	NULL,			/* get_xfont_device */
+	NULL,			/* map_rgb_alpha_color */
+	gx_page_device_get_page_device,
+	win_get_alpha_bits
 };
-#ifdef __DLL__
-gx_device_win_dib gs_mswindll_device = {
+gx_device_win_dib far_data gs_mswindll_device = {
 	std_device_std_body(gx_device_win_dib, &win_dib_procs, "mswindll",
 	  INITIAL_WIDTH, INITIAL_HEIGHT, 	/* win_open() fills these in later */
 	  INITIAL_RESOLUTION, INITIAL_RESOLUTION	/* win_open() fills these in later */
 	),
 	 { 0 },				/* std_procs */
 	0,				/* BitsPerPixel */
-	5000,				/* UpdateInterval (in milliseconds) */
-	"\0",				/* GSVIEW_STR */
-	1,				/* This is a DLL device */
 	2,				/* nColors */
+	1,				/* Text Alpha bits */
+	1,				/* Graphics Alpha bits */
 	0,				/* mapped_color_flags */
-	win_dib_copy_to_clipboard,
-	win_dib_repaint,
-	win_dib_alloc_bitmap,
-	win_dib_free_bitmap
-};
-#endif
-gx_device_win_dib gs_mswin_device = {
-	std_device_std_body(gx_device_win_dib, &win_dib_procs, "mswin",
-	  INITIAL_WIDTH, INITIAL_HEIGHT, 	/* win_open() fills these in later */
-	  INITIAL_RESOLUTION, INITIAL_RESOLUTION	/* win_open() fills these in later */
-	),
-	 { 0 },				/* std_procs */
-	0,				/* BitsPerPixel */
-	5000,				/* UpdateInterval (in milliseconds) */
-	"\0",				/* GSVIEW_STR */
-	0,				/* not a DLL device */
-	2,				/* nColors */
-	0,				/* mapped_color_flags */
-	win_dib_copy_to_clipboard,
-	win_dib_repaint,
 	win_dib_alloc_bitmap,
 	win_dib_free_bitmap
 };
 
 /* forward declarations */
 private HGLOBAL win_dib_make_dib(gx_device_win *dev, int orgx, int orgy, int wx, int wy);
+private int win_dib_lock_device(unsigned char *device, int flag);
 
 
 /* Open the win_dib driver */
@@ -130,19 +127,24 @@ win_dib_open(gx_device *dev)
 	int code = win_open(dev);
 	if ( code < 0 ) return code;
 
+#ifdef __WIN32__
+	if (!is_win32s)
+	    wdev->hmtx = CreateMutex(NULL, FALSE, NULL); /* unnamed mutex, initially unowned */
+#endif
 	if ( gdev_mem_device_for_bits(dev->color_info.depth) == 0 )
 	{	win_close(dev);
 		return gs_error_rangecheck;
 	}
 	code = win_dib_alloc_bitmap((gx_device_win *)dev, dev);
-#ifdef __DLL__
-	if ( code < 0 ) return code;
-	if (wdev->dll) {
-		/* notify caller about new device */
-		(*pgsdll_callback)(GSDLL_DEVICE, (unsigned char *)dev, 1);
-		return 0;	/* caller will handle displaying */
+	if ( code < 0 ) {
+	     win_close(dev);
+	     return code;
 	}
-#endif
+	/* notify caller about new device */
+	(*pgsdll_callback)(GSDLL_DEVICE, (unsigned char *)dev, 1);
+	(*pgsdll_callback)(GSDLL_SIZE, (unsigned char *)dev,
+		     (dev->width & 0xffff) +
+		     ((ulong)(dev->height & 0xffff) << 16));
 	return code;
 }
 
@@ -162,13 +164,16 @@ win_dib_get_initial_matrix(gx_device *dev, gs_matrix *pmat)
 private int
 win_dib_close(gx_device *dev)
 {int code;
+	/* wait until bitmap is not being used by caller */
+        win_dib_lock_device((unsigned char *)dev, 1);
+	(*pgsdll_callback)(GSDLL_DEVICE, (unsigned char *)dev, 0);
+        win_dib_lock_device((unsigned char *)dev, 0);
 	win_dib_free_bitmap((gx_device_win *)dev);
-	code = win_close(dev);
-#ifdef __DLL__
-        if (wdev->dll) {
-            (*pgsdll_callback)(GSDLL_DEVICE, (unsigned char *)dev, 0);
-        }
+#ifdef __WIN32__
+	if (!is_win32s)
+	    CloseHandle(wdev->hmtx);
 #endif
+	code = win_close(dev);
 	return code;
 }
 
@@ -212,7 +217,6 @@ win_dib_fill_rectangle(gx_device *dev, int x, int y, int w, int h,
 #else
 	wmproc(fill_rectangle)(wmdev, x, y, w, h, color);
 #endif
-	win_update((gx_device_win *)dev);
 	return 0;
 }
 
@@ -243,7 +247,6 @@ win_dib_copy_mono(gx_device *dev,
 	wmproc(copy_mono)(wmdev, base, sourcex, raster, id,
 			  x, y, w, h, zero, one);
 #endif
-	win_update((gx_device_win *)dev);
 	return 0;
 }
 
@@ -272,19 +275,36 @@ win_dib_copy_color(gx_device *dev,
 	wmproc(copy_color)(wmdev, base, sourcex, raster, id,
 			   x, y, w, h);
 #endif
-	win_update((gx_device_win *)dev);
 	return 0;
 }
 
+int
+win_dib_get_bits(gx_device *dev, int y, byte *str, byte **actual_data)
+{
+	return wmproc(get_bits)(wmdev, y, str, actual_data);
+}
+
+int
+win_dib_put_params(gx_device *dev, gs_param_list *plist)
+{
+int code;
+    win_dib_lock_device((unsigned char *)dev, 1);
+    code = win_put_params(dev, plist);
+    win_dib_lock_device((unsigned char *)dev, 0);
+    return code;
+}
+
 /* ------ DLL device procedures ------ */
-#ifdef __DLL__
+
 /* make a copy of the device bitmap and return shared memory handle to it */
 /* device is a pointer to Ghostscript device from GSDLL_DEVICE message */
 HGLOBAL WINAPI
 gsdll_copy_dib(unsigned char *device)
 {
-gx_device_win *dev = (gx_device_win *)device;
-	return win_dib_make_dib(dev, 0, 0, wdev->width, wdev->height);
+gx_device_win_dib *dev = (gx_device_win_dib *)device;
+	if (!dev || !dev->is_open || dev->mdev.width==0 || dev->mdev.height==0)
+	    return (HGLOBAL)NULL;
+	return win_dib_make_dib((gx_device_win *)dev, 0, 0, dev->width, dev->height);
 }
 
 /* make a copy of the device palette and return a handle to it */
@@ -292,8 +312,12 @@ gx_device_win *dev = (gx_device_win *)device;
 HPALETTE WINAPI
 gsdll_copy_palette(unsigned char *device)
 {
-gx_device_win *dev = (gx_device_win *)device;
-	return CreatePalette(wdev->limgpalette);
+gx_device_win_dib *dev = (gx_device_win_dib *)device;
+	if (!dev || !dev->is_open || dev->mdev.width==0 || dev->mdev.height==0)
+	    return (HPALETTE)NULL;
+	if (wdev->nColors > 0)
+	    return CreatePalette(dev->limgpalette);
+	return (HPALETTE)NULL;
 }
 
 /* copy the rectangle src from the device bitmap */
@@ -303,21 +327,22 @@ gx_device_win *dev = (gx_device_win *)device;
 void WINAPI
 gsdll_draw(unsigned char *device, HDC hdc, LPRECT dest, LPRECT src)
 {
-gx_device_win *dev = (gx_device_win *)device;
+gx_device_win_dib *dev = (gx_device_win_dib *)device;
 HPALETTE oldpalette;
-	if (wdev->nColors > 0) {
-	    oldpalette = SelectPalette(hdc,wdev->himgpalette,NULL);
+	if (!dev || !dev->is_open || dev->mdev.width==0 || dev->mdev.height==0)
+	    return;
+	if (dev->nColors > 0) {
+	    oldpalette = SelectPalette(hdc,dev->himgpalette,NULL);
 	    RealizePalette(hdc);
 	}
-	win_dib_repaint(wdev,hdc, src->left, src->top,
-	    src->right-src->left, src->bottom-src->top,  
-	    dest->left, dest->top);
-	if (wdev->nColors > 0) {
+	win_dib_repaint((gx_device_win *)dev, hdc, dest->left, dest->top,
+	    dest->right-dest->left, dest->bottom-dest->top,  
+	    src->left, src->top);
+	if (dev->nColors > 0) {
 	    SelectPalette(hdc,oldpalette,NULL);
 	}
 	return;
 }
-#endif /* __DLL__ */
 
 /* ------ Windows-specific device procedures ------ */
 
@@ -327,10 +352,11 @@ private void
 win_dib_repaint(gx_device_win *dev, HDC hdc, int dx, int dy, int wx, int wy,
   int sx, int sy)
 {	struct bmi_s {
-		BITMAPINFOHEADER h;
-		ushort pal_index[256];
+	  BITMAPINFOHEADER h;
+	  ushort pal_index[256];
 	} bmi;
 	int i;
+	UINT which_colors;
 
 	bmi.h.biSize = sizeof(bmi.h);
 	bmi.h.biWidth = wdev->mdev.width;
@@ -341,27 +367,39 @@ win_dib_repaint(gx_device_win *dev, HDC hdc, int dx, int dy, int wx, int wy,
 	bmi.h.biSizeImage = 0;			/* default */
 	bmi.h.biXPelsPerMeter = 0;		/* default */
 	bmi.h.biYPelsPerMeter = 0;		/* default */
-	if (dev->BitsPerPixel <= 8) {
-	    bmi.h.biClrUsed = wdev->nColors;
-	    bmi.h.biClrImportant = wdev->nColors;
-	    for ( i = 0; i < wdev->nColors; i++ )
-		bmi.pal_index[i] = i;
-	    SetDIBitsToDevice(hdc, dx, dy, wx, wy,
-		sx, 0, 0, wy,
-		wdev->mdev.line_ptrs[wdev->height - (sy + wy)],
-		(BITMAPINFO FAR *)&bmi, DIB_PAL_COLORS);
-	}
-	else {
-	    bmi.h.biClrUsed = 0;
-	    bmi.h.biClrImportant = 0;
-	    SetDIBitsToDevice(hdc, dx, dy, wx, wy,
-		sx, 0, 0, wy,
-		wdev->mdev.line_ptrs[wdev->height - (sy + wy)],
-		(BITMAPINFO FAR *)&bmi, DIB_RGB_COLORS);
+	if ( dev->BitsPerPixel <= 8 )
+	  {	bmi.h.biClrUsed = wdev->nColors;
+		bmi.h.biClrImportant = wdev->nColors;
+		for ( i = 0; i < wdev->nColors; i++ )
+		  bmi.pal_index[i] = i;
+		which_colors = DIB_PAL_COLORS;
+	  }
+	else
+	  {	bmi.h.biClrUsed = 0;
+		bmi.h.biClrImportant = 0;
+		which_colors = DIB_RGB_COLORS;
+	  }
+	/*
+	 * Windows apparently limits the size of a single transfer
+	 * to 2 Mb, which can be exceeded on 24-bit displays.
+	 * Deal with this here.
+	 */
+#define max_transfer 2000000
+	if ( wdev->mdev.raster > 0 )		/* just in case! */
+	{	long ny = max_transfer / wdev->mdev.raster;
+		for ( ; wy > ny; dy += ny, wy -= ny, sy += ny )
+		  SetDIBitsToDevice(hdc, dx, dy, wx, ny,
+			sx, 0, 0, ny,
+			wdev->mdev.line_ptrs[wdev->height - (sy + ny)],
+			(BITMAPINFO FAR *)&bmi, which_colors);
 	}
+#undef max_transfer
+	SetDIBitsToDevice(hdc, dx, dy, wx, wy,
+			  sx, 0, 0, wy,
+			  wdev->mdev.line_ptrs[wdev->height - (sy + wy)],
+			  (BITMAPINFO FAR *)&bmi, which_colors);
 }
 
-
 /* This makes a DIB that contains all or part of the bitmap. */
 /* The bitmap pixel orgx must start on a byte boundary. */
 private HGLOBAL
@@ -376,11 +414,13 @@ win_dib_make_dib(gx_device_win *dev, int orgx, int orgy, int wx, int wy)
 	BYTE huge *pBits;
 	BYTE huge *pLine;
 	ulong bitmapsize;
-	int i;
-	int loffset; 		/* byte offset to start of line */
-	UINT lwidth;		/* line width in bytes rounded up to multiple of 4 bytes */
-	UINT lseg;		/* bytes remaining in this segment */
 	int palcount;
+	int i;
+	UINT lwidth;		/* line width in bytes rounded up to multiple of 4 bytes */
+#ifdef USE_SEGMENTS
+	int loffset; 		/* byte offset to start of line */
+	UINT lseg;		/* bytes remaining in this segment */
+#endif
 
 	if (orgx + wx > wdev->width)
 		wx = wdev->width - orgx;
@@ -453,26 +493,6 @@ win_dib_make_dib(gx_device_win *dev, int orgx, int orgy, int wx, int wy)
 }
 
 
-/* Copy the bitmap to the clipboard. */
-private void
-win_dib_copy_to_clipboard(gx_device_win *dev)
-{
-	HGLOBAL hglobal;
-	hglobal = win_dib_make_dib(dev, 0, 0, wdev->width, wdev->height);
-	if (hglobal == (HGLOBAL)NULL) {
-	    MessageBox(wdev->hwndimg, "Not enough memory to Copy to Clipboard", 
-		szAppName, MB_OK | MB_ICONEXCLAMATION);
-	    return;
-	}
-	OpenClipboard(dev->hwndimg);
-	EmptyClipboard();
-	SetClipboardData(CF_DIB, hglobal);
-	if (wdev->nColors > 0)
-	    SetClipboardData(CF_PALETTE, CreatePalette(wdev->limgpalette));
-	CloseClipboard();
-}
-
-
 /* Allocate the backing bitmap. */
 private int
 win_dib_alloc_bitmap(gx_device_win *dev, gx_device *param_dev)
@@ -482,21 +502,32 @@ win_dib_alloc_bitmap(gx_device_win *dev, gx_device *param_dev)
 	HGLOBAL hmdata;
 	byte huge *base;
 	byte huge *ptr_base;
-	ulong data_size;
 	uint ptr_size;
 	uint raster;
+#ifdef USE_SEGMENTS
+	ulong data_size;
+#endif
 
-	/* Round up the width so that the scan line size is a power of 2. */
-	if (dev->color_info.depth == 24) {
-	    width = param_dev->width * 3 - 1;
-	    while ( width & (width + 1) ) width |= width >> 1;
-	    width = (width + 1) / 3;
+#ifdef __WIN32__
+	if (is_win32s) {
+#endif
+	  /* Round up the width so that the scan line size is a power of 2. */
+	    if (dev->color_info.depth == 24) {
+		width = param_dev->width * 3 - 1;
+		while ( width & (width + 1) ) width |= width >> 1;
+		width = (width + 1) / 3;
+	    }
+	    else {
+		width = param_dev->width - 1;
+		while ( width & (width + 1) ) width |= width >> 1;
+		width++;
+	    }
+#ifdef __WIN32__
 	}
-	else {
-	    width = param_dev->width - 1;
-	    while ( width & (width + 1) ) width |= width >> 1;
-	    width++;
+	else {	/* don't have to worry about segments so use less memory */
+	    width = param_dev->width;
 	}
+#endif
 
 	/* Finish initializing the DIB. */
 
@@ -507,8 +538,9 @@ win_dib_alloc_bitmap(gx_device_win *dev, gx_device *param_dev)
 	data_size = (ulong)raster * mdev.height;
 	ptr_size = sizeof(byte **) * mdev.height;
 	hmdata = GlobalAlloc(0, raster + data_size + ptr_size * 2);
-	if ( hmdata == 0 )
+	if ( hmdata == 0 ) {
 		return win_nomemory();
+	}
 
 	/* Nothing can go wrong now.... */
 
@@ -529,6 +561,12 @@ win_dib_alloc_bitmap(gx_device_win *dev, gx_device *param_dev)
 	wdev->mdev = mdev;
 	wdev->mdev.base = (byte *)base;
 	wmproc(open_device)((gx_device *)&wdev->mdev);
+
+	if (wdev->is_open)
+	  (*pgsdll_callback)(GSDLL_SIZE, (unsigned char *)dev,
+			     (dev->width & 0xffff) +
+			     ((ulong)(dev->height & 0xffff) << 16));
+
 	return 0;
 }
 
@@ -540,3 +578,102 @@ win_dib_free_bitmap(gx_device_win *dev)
 	GlobalUnlock(hmdata);
 	GlobalFree(hmdata);
 }
+
+/* Lock the device (so it's size cannot be changed) if flag = TRUE */
+/* or unlock the device if flag = FALSE */
+/* device is a pointer to Ghostscript device from GSDLL_DEVICE message */
+private int
+win_dib_lock_device(unsigned char *device, int flag)
+{
+gx_device *dev = (gx_device *)device;
+#ifdef __WIN32__
+    if (!is_win32s) {
+	if (flag) {
+	    if (WaitForSingleObject(wdev->hmtx, 60000) == WAIT_TIMEOUT)
+		return 2;
+	    return 1;
+	}
+	ReleaseMutex(wdev->hmtx);
+	return 0;
+    }
+#endif
+    if (flag)
+	wdev->lock_count++;
+    else 
+	wdev->lock_count--;
+    if (wdev->lock_count < 0)
+	wdev->lock_count = 0;
+    return wdev->lock_count;
+}
+
+int GSDLLAPI _export
+gsdll_lock_device(unsigned char *device, int flag)
+{
+    return win_dib_lock_device(device, flag);
+}
+
+
+/* Copy bitmap
+ * If pbmih nonzero, copy the BITMAPINFOHEADER.
+ * If prgbquad nonzero, copy the palette.
+ *   number of entries copied is given by pbmih->biClrUsed
+ * If ppbyte nonzero, return pointer to row.
+ *   pointer is only valid while device is locked
+ * GS can change the palette while the device is locked.
+ * Do not call this function while GS is busy.
+ * If all pbmih and prgbquad and ppbyte are all NULL,
+ * return value is byte count needed for BITMAPINFOHEADER
+ * and palette and one bitmap row.
+ * Otherwise return value is 0;
+ *
+ * This function exists to allow the bitmap to be copied to a file
+ * or structured storage, without the overhead of having two copies
+ * of the bitmap in memory at the same time.
+ */
+int GSDLLAPI _export
+gsdll_get_bitmap_row(unsigned char *device, LPBITMAPINFOHEADER pbmih,
+    LPRGBQUAD prgbquad, LPBYTE *ppbyte, unsigned int row)
+{
+    int palcount;
+    gx_device_win_dib *dev = (gx_device_win_dib *)device;
+    palcount = (dev->color_info.depth == 24) ? 0 : dev->nColors;
+
+    if (pbmih) {
+	pbmih->biSize = sizeof(BITMAPINFOHEADER);
+	pbmih->biWidth = dev->width;
+	pbmih->biHeight = dev->mdev.height;
+	pbmih->biPlanes = 1;
+	pbmih->biBitCount = dev->color_info.depth;
+	pbmih->biCompression = 0;
+	pbmih->biSizeImage = 0;			/* default */
+	pbmih->biXPelsPerMeter = (DWORD)(dev->x_pixels_per_inch / 25.4 * 1000);
+	pbmih->biYPelsPerMeter = (DWORD)(dev->y_pixels_per_inch / 25.4 * 1000);
+	pbmih->biClrUsed = palcount;
+	pbmih->biClrImportant = palcount;
+    }
+
+    if (prgbquad) {
+	int i;
+        gx_color_value prgb[3];
+	for ( i = 0; i < palcount; i++ ) {
+	    win_map_color_rgb((gx_device *)wdev, (gx_color_index)i, prgb);
+	    prgbquad[i].rgbRed   = win_color_value(prgb[0]);
+	    prgbquad[i].rgbGreen = win_color_value(prgb[1]);
+	    prgbquad[i].rgbBlue  = win_color_value(prgb[2]);
+	    prgbquad[i].rgbReserved = 0;
+	}
+    }
+
+    if (ppbyte) {
+	if (row < dev->mdev.height)
+	    *ppbyte = dev->mdev.line_ptrs[row];
+	else
+	    *ppbyte = NULL;
+    }
+
+    if ((pbmih == NULL) && (prgbquad == NULL) && (ppbyte == NULL))
+	return sizeof(BITMAPINFOHEADER) + 256 * sizeof(RGBQUAD)
+	    + gdev_mem_raster(&(dev->mdev));
+    return 0;
+}
+
