@@ -1,13 +1,24 @@
-/* Copyright (C) 1989, 1992, 1993, 1994 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevxini.c */
@@ -178,11 +189,13 @@ gdev_x_open(register gx_device_X *xdev)
     int xid_height, xid_width;
 
 #ifdef DEBUG
+# ifdef have_Xdebug
     if (gs_debug['X']) {
 	extern int _Xdebug;
 
 	_Xdebug = 1;
     }
+# endif
 #endif
     if (!(xdev->dpy = XOpenDisplay((char *)NULL))) {
 	char *dispname = getenv("DISPLAY");
@@ -374,10 +387,8 @@ gdev_x_open(register gx_device_X *xdev)
 	/* Take care of resolution and paper size. */
 	if (xdev->x_pixels_per_inch == FAKE_RES ||
 	    xdev->y_pixels_per_inch == FAKE_RES) {
-	    float xsize, ysize;
-
-	    xsize = (float)xdev->width / xdev->x_pixels_per_inch;
-	    ysize = (float)xdev->height / xdev->y_pixels_per_inch;
+	    float xsize = (float)xdev->width / xdev->x_pixels_per_inch;
+	    float ysize = (float)xdev->height / xdev->y_pixels_per_inch;
 
 	    if (xdev->xResolution == 0.0 && xdev->yResolution == 0.0) {
 		float dpi, xdpi, ydpi;
@@ -387,11 +398,21 @@ gdev_x_open(register gx_device_X *xdev)
 		ydpi = 25.4 * HeightOfScreen(xdev->scr) /
 		       HeightMMOfScreen(xdev->scr);
 		dpi = min(xdpi, ydpi);
-
-		while (xsize*dpi > WidthOfScreen(xdev->scr)-32 ||
-		       ysize*dpi > HeightOfScreen(xdev->scr)-32) {
-		    dpi *= 0.95;
-		}
+		/*
+		 * Some X servers provide very large "virtual screens", and
+		 * return the virtual screen size for Width/HeightMM but the
+		 * physical size for Width/Height.  Attempt to detect and
+		 * correct for this now.  This is a KLUDGE required because
+		 * the X server provides no way to read the screen
+		 * resolution directly.
+		 */
+		if ( dpi < 30 )
+		  dpi = 75;		/* arbitrary */
+		else
+		  { while (xsize*dpi > WidthOfScreen(xdev->scr)-32 ||
+			   ysize*dpi > HeightOfScreen(xdev->scr)-32)
+		      dpi *= 0.95;
+		  }
 		xdev->x_pixels_per_inch = dpi;
 		xdev->y_pixels_per_inch = dpi;
 	    } else {
@@ -404,6 +425,10 @@ gdev_x_open(register gx_device_X *xdev)
 	    if (xdev->height > HeightOfScreen(xdev->scr)) {
 		xdev->height = ysize*xdev->y_pixels_per_inch;
 	    }
+	    xdev->MediaSize[0] =
+	      (float)xdev->width / xdev->x_pixels_per_inch * 72;
+	    xdev->MediaSize[1] =
+	      (float)xdev->height / xdev->y_pixels_per_inch * 72;
 	}
 
 	sizehints.x = 0;
@@ -559,7 +584,9 @@ gdev_x_clear_window(gx_device_X *xdev)
 	    XSync(xdev->dpy, False);	/* Force the error */
 	    if (alloc_error) {
 		xdev->useBackingPixmap = False;
+#ifdef DEBUG
 		eprintf("Warning: Failed to allocated backing pixmap.\n");
+#endif
 		if (xdev->bpixmap) {
 		    XFreePixmap(xdev->dpy, xdev->bpixmap);
 		    xdev->bpixmap = None;
@@ -783,10 +810,12 @@ gdev_x_setup_colors(gx_device_X *xdev)
 		xdev->color_info.dither_grays =
 		    xdev->color_info.dither_colors = ramp_size;
 		if ( !setup_cube(xdev, ramp_size, true) ) {
-		    /* eprintf3("Warning: failed to allocate %dx%dx%d RGB cube.\n",
-			     ramp_size, ramp_size, ramp_size); */
-		    ramp_size--;
-		    continue;
+#ifdef DEBUG
+		  eprintf3("Warning: failed to allocate %dx%dx%d RGB cube.\n",
+			   ramp_size, ramp_size, ramp_size);
+#endif
+		  ramp_size--;
+		  continue;
 		}
 	    }
 
@@ -824,10 +853,13 @@ grayscale:
 	    while (!xdev->dither_colors && ramp_size >= 3) {
 		xdev->color_info.dither_grays = ramp_size;
 		if ( !setup_cube(xdev, ramp_size, false) )
-		{   /* eprintf1("Warning: failed to allocate %d level gray ramp.\n",
-			     ramp_size); */
-		    ramp_size /= 2;
-		    continue;
+		{
+#ifdef DEBUG
+		  eprintf1("Warning: failed to allocate %d level gray ramp.\n",
+			   ramp_size);
+#endif
+		  ramp_size /= 2;
+		  continue;
 		}
 	    }
 	    if (!xdev->dither_colors) {
