@@ -1,13 +1,24 @@
-/* Copyright (C) 1989, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gschar.c */
@@ -22,11 +33,13 @@
 #include "gxmatrix.h"
 #include "gzstate.h"
 #include "gxcoord.h"
+#include "gxdevice.h"
 #include "gxdevmem.h"
 #include "gxchar.h"
 #include "gxfont.h"
 #include "gxfont0.h"
 #include "gxfcache.h"
+#include "gspath.h"
 #include "gzpath.h"
 
 /* Define whether or not to cache characters rotated by angles other than */
@@ -36,25 +49,31 @@
 /* Define whether or not to oversample characters at small sizes. */
 #define OVERSAMPLE 1
 
+/* Define the maximum size of a full temporary bitmap when rasterizing, */
+/* in bits (not bytes). */
+#define MAX_TEMP_BITMAP_BITS 80000
+
 /* Structure descriptors */
 private_st_gs_show_enum();
 #define eptr ((gs_show_enum *)vptr)
 private ENUM_PTRS_BEGIN(show_enum_enum_ptrs) {
-	if ( index > eptr->fstack.depth + 4 )
+	if ( index > eptr->fstack.depth + 5 )
 	  return 0;
-	*pep = eptr->fstack.items[index - 4].font;
+	*pep = eptr->fstack.items[index - 5].font;
 	break;
 	}
 	ENUM_PTR(0, gs_show_enum, pgs);
 	ENUM_CONST_STRING_PTR(1, gs_show_enum, str);
 	ENUM_PTR(2, gs_show_enum, dev_cache);
-	ENUM_PTR(3, gs_show_enum, dev_null);
+	ENUM_PTR(3, gs_show_enum, dev_cache2);
+	ENUM_PTR(4, gs_show_enum, dev_null);
 ENUM_PTRS_END
 private RELOC_PTRS_BEGIN(show_enum_reloc_ptrs) {
 	int i;
 	RELOC_PTR(gs_show_enum, pgs);
 	RELOC_CONST_STRING_PTR(gs_show_enum, str);
 	RELOC_PTR(gs_show_enum, dev_cache);
+	RELOC_PTR(gs_show_enum, dev_cache2);
 	RELOC_PTR(gs_show_enum, dev_null);
 	for ( i = 0; i <= eptr->fstack.depth; i++ )
 	  RELOC_PTR(gs_show_enum, fstack.items[i].font);
@@ -69,7 +88,7 @@ private int show_setup(P3(gs_show_enum *, gs_state *, const char *));
 private void show_set_scale(P1(gs_show_enum *));
 private int show_cache_setup(P1(gs_show_enum *));
 private int show_state_setup(P1(gs_show_enum *));
-private int show_origin_setup(P4(gs_state *, fixed, fixed, int));
+private int show_origin_setup(P4(gs_state *, fixed, fixed, gs_char_path_mode));
 private int stringwidth_setup(P3(gs_show_enum *, gs_state *, const char *));
 
 /* Print the ctm if debugging */
@@ -108,6 +127,7 @@ gs_show_enum_alloc(gs_memory_t *mem, gs_state *pgs, client_name_t cname)
 	penum->str.data = 0;
 	penum->str.size = 0;
 	penum->dev_cache = 0;
+	penum->dev_cache2 = 0;
 	penum->dev_null = 0;
 	penum->fstack.depth = -1;
 	return penum;
@@ -118,6 +138,13 @@ void
 gs_show_enum_release(gs_show_enum *penum, gs_memory_t *emem)
 {	gs_state *pgs = penum->pgs;
 	gs_memory_t *mem = pgs->memory;
+
+	penum->cc = 0;
+	if ( penum->dev_cache2 != 0 )
+	  {	gs_free_object(mem, penum->dev_cache2,
+			       "gs_show_enum_release(dev_cache2)");
+		penum->dev_cache2 = 0;
+	  }
 	if ( penum->dev_cache != 0 )
 	  {	gs_free_object(mem, penum->dev_cache,
 			       "gs_show_enum_release(dev_cache)");
@@ -313,6 +340,7 @@ stringwidth_setup(gs_show_enum *penum, gs_state *pgs, const char *str)
 	gs_make_null_device(dev_null, mem);
 	dev_null->target = gs_currentdevice_inline(pgs);
 	pgs->device = (gx_device *)dev_null;
+	pgs->ctm_default_set = false;
 	penum->dev_null = dev_null;
 	/* Establish an arbitrary translation and current point. */
 	gs_newpath(pgs);
@@ -323,21 +351,42 @@ stringwidth_setup(gs_show_enum *penum, gs_state *pgs, const char *str)
 /* charpath[_n] */
 int
 gs_charpath_n_init(gs_show_enum *penum, gs_state *pgs,
-  const char *str, uint size, int stroke_path)
+  const char *str, uint size, bool stroke_path)
 {	int code;
 	setup_show_n();
 	code = show_setup(penum, pgs, str);
-	penum->charpath_flag = (stroke_path ? 2 : 1);
-	if_debug0('k', "[k]charpath, can_cache=false");
-	penum->can_cache = false;
+	penum->charpath_flag =
+	  (stroke_path ? cpm_true_charpath : cpm_false_charpath);
+	penum->can_cache = -1;
+	if_debug1('k', "[k]charpath, can_cache=%d", penum->can_cache);
 	return code;
 }
 int
 gs_charpath_init(gs_show_enum *penum, gs_state *pgs,
-  const char *str, int stroke_path)
+  const char *str, bool stroke_path)
 {	return gs_charpath_n_init(penum, pgs, str, strlen(str), stroke_path);
 }
 
+/* charboxpath[_n] */
+int
+gs_charboxpath_n_init(gs_show_enum *penum, gs_state *pgs,
+  const char *str, uint size, bool use_boxes)
+{	int code;
+	setup_show_n();
+	code = show_setup(penum, pgs, str);
+	penum->charpath_flag =
+	  (use_boxes ? cpm_true_charboxpath : cpm_false_charboxpath);
+	penum->can_cache = 0;		/* different from charpath! */
+	if_debug1('k', "[k]charboxpath, can_cache=%d", penum->can_cache);
+	return code;
+}
+int
+gs_charboxpath_init(gs_show_enum *penum, gs_state *pgs,
+  const char *str, bool stroke_path)
+{	return gs_charboxpath_n_init(penum, pgs, str, strlen(str),
+				     stroke_path);
+}
+
 /* ------ Width/cache operators ------ */
 
 private int set_cache_device(P3(gs_show_enum *, gs_state *,
@@ -345,48 +394,49 @@ private int set_cache_device(P3(gs_show_enum *, gs_state *,
 
 /* setcachedevice */
 /* The elements of pw are: wx, wy, llx, lly, urx, ury. */
+/* Note that this returns 1 if we just set up the cache device. */
 int
 gs_setcachedevice(gs_show_enum *penum, gs_state *pgs, const float *pw)
-{	int wcode = gs_setcharwidth(penum, pgs, pw[0], pw[1]);	/* default is don't cache */
-	int code;
-	if ( wcode < 0 )
-	  return wcode;
-	code = set_cache_device(penum, pgs, pw + 2);
-	return (code < 0 ? code : wcode);
+{	int code = gs_setcharwidth(penum, pgs, pw[0], pw[1]);	/* default is don't cache */
+
+	if ( code < 0 )
+	  return code;
+	return set_cache_device(penum, pgs, pw + 2);
 }
 
 /* setcachedevice2 */
 /* The elements of pw2 are: w0x, w0y, llx, lly, urx, ury, w1x, w1y, vx, vy. */
+/* Note that this returns 1 if we just set up the cache device. */
 int
 gs_setcachedevice2(gs_show_enum *penum, gs_state *pgs, const float *pw2)
-{	int wcode, code;
+{	int code;
+
 	if ( gs_rootfont(pgs)->WMode )
 	{	float vx = pw2[8], vy = pw2[9];
 		gs_fixed_point pvxy, dvxy;
 		cached_char *cc;
 		if ( (code = gs_point_transform2fixed(&pgs->ctm, -vx, -vy, &pvxy)) < 0 ||
 		     (code = gs_distance_transform2fixed(&pgs->ctm, vx, vy, &dvxy)) < 0 ||
-		     (code = wcode = gs_setcharwidth(penum, pgs, pw2[6], pw2[7])) < 0
+		     (code = gs_setcharwidth(penum, pgs, pw2[6], pw2[7])) < 0
 		   )
 		  return code;
 		/* Adjust the origin by (vx, vy). */
 		gx_translate_to_fixed(pgs, pvxy.x, pvxy.y);
 		code = set_cache_device(penum, pgs, pw2 + 2);
 		if ( code != 1 )
-		  return (code < 0 ? code : wcode);
+		  return code;
 		/* Adjust the character origin too. */
 		cc = penum->cc;
 		cc->offset.x += dvxy.x;
 		cc->offset.y += dvxy.y;
-		return wcode;
 	}
 	else
-	{	wcode = gs_setcharwidth(penum, pgs, pw2[0], pw2[1]);
-		if ( wcode < 0 )
-		  return wcode;
+	{	code = gs_setcharwidth(penum, pgs, pw2[0], pw2[1]);
+		if ( code < 0 )
+		  return code;
 		code = set_cache_device(penum, pgs, pw2 + 2);
-		return (code < 0 ? code : wcode);
 	}
+	return code;
 }
 
 /* Set up the cache device if relevant. */
@@ -407,9 +457,9 @@ set_cache_device(register gs_show_enum *penum, gs_state *pgs, const float *pb)
 	  return 0;
 	/* We can only use the cache if ctm is unchanged */
 	/* (aside from a possible translation). */
-	if ( !penum->can_cache || !pgs->char_tm_valid )
+	if ( penum->can_cache <= 0 || !pgs->char_tm_valid )
 	{	if_debug2('k', "[k]no cache: can_cache=%d, char_tm_valid=%d\n",
-			  (int)penum->can_cache, (int)pgs->char_tm_valid);
+			  penum->can_cache, (int)pgs->char_tm_valid);
 		return 0;
 	}
 	   {	const gs_font *pfont = pgs->font;
@@ -479,8 +529,14 @@ if ( gs_debug_c('k') )
 		  {	int more_bits =
 			  alpha_bits - (log2_scale.x + log2_scale.y);
 			if ( more_bits > 0 )
-			  {	log2_scale.x += (more_bits + 1) >> 1;
-				log2_scale.y += more_bits >> 1;
+			  {    if (log2_scale.x <= log2_scale.y) 
+				  { log2_scale.x += (more_bits + 1) >> 1;
+				    log2_scale.y += more_bits >> 1;
+				  }
+				else
+				  { log2_scale.x += more_bits >> 1;
+				    log2_scale.y += (more_bits + 1) >> 1;
+				  }
 			  }
 		  }
 		else
@@ -504,8 +560,18 @@ if ( gs_debug_c('k') )
 			if ( code < 0 )
 			  return code;
 		  }
+		/*
+		 * If we're oversampling (i.e., the temporary bitmap is
+		 * larger than the final monobit or alpha array) and the
+		 * temporary bitmap is large, use incremental conversion
+		 * from oversampled bitmap strips to alpha values instead of
+		 * full oversampling with compression at the end.
+		 */
 		cc = gx_alloc_char_bits(dir, penum->dev_cache,
-					iwidth, iheight, &log2_scale);
+			(iwidth > MAX_TEMP_BITMAP_BITS / iheight &&
+			 (1 << (log2_scale.x + log2_scale.y)) > alpha_bits ?
+			 penum->dev_cache2 : NULL),
+			iwidth, iheight, &log2_scale, alpha_bits);
 		if ( cc == 0 )
 		  return 0;	/* too big for cache */
 		/* The mins handle transposed coordinate systems.... */
@@ -524,10 +590,10 @@ if ( gs_debug_c('k') )
 		penum->cc = cc;
 		cc->code = glyph;
 		cc->wmode = gs_rootfont(pgs)->WMode;
-		cc->depth = alpha_bits;
 		cc->wxy = penum->wxy;
 		/* Install the device */
 		pgs->device = (gx_device *)penum->dev_cache;
+		pgs->ctm_default_set = false;
 		/* Adjust the transformation in the graphics context */
 		/* so that the character lines up with the cache. */
 		gx_translate_to_fixed(pgs,
@@ -555,6 +621,8 @@ if ( gs_debug_c('k') )
 }
 
 /* setcharwidth */
+/* Note that this returns 1 if the current show operation is */
+/* non-displaying (stringwidth or cshow). */
 int
 gs_setcharwidth(register gs_show_enum *penum, gs_state *pgs,
   floatp wx, floatp wy)
@@ -647,8 +715,10 @@ show_update(register gs_show_enum *penum)
 		gx_add_cached_char(pgs->font->dir, penum->dev_cache,
 				   cc, gx_lookup_fm_pair(pgs->font, pgs),
 				   &penum->log2_current_scale);
-		if ( penum->stringwidth_flag || penum->charpath_flag )
-			break;
+		if ( penum->stringwidth_flag != 0 ||
+		     penum->charpath_flag != cpm_show
+		   )
+		  break;
 		/* falls through */
 	case sws_cache_width_only:
 		/* Copy the bits to the real output device. */
@@ -662,7 +732,7 @@ show_update(register gs_show_enum *penum)
 	case sws_no_cache:
 		;
 	   }
-	if ( penum->charpath_flag )
+	if ( penum->charpath_flag != cpm_show )
 	{	/* Move back to the character origin, so that */
 		/* show_move will get us to the right place. */
 		code = gx_path_add_point(pgs->show_gstate->path,
@@ -731,7 +801,7 @@ show_proceed(register gs_show_enum *penum)
 	gx_device *dev = gs_currentdevice_inline(pgs);
 	int alpha_bits = (*dev_proc(dev, get_alpha_bits))(dev, go_text);
 
-	if ( !(penum->charpath_flag | penum->stringwidth_flag) )
+	if ( penum->charpath_flag == cpm_show && !penum->stringwidth_flag )
 	  {	code = gx_color_load(pgs->dev_color, pgs);
 		if ( code < 0 )
 		  return code;
@@ -739,10 +809,10 @@ show_proceed(register gs_show_enum *penum)
 more:	/* Proceed to next character */
 	pfont = (penum->fstack.depth < 0 ? pgs->font :
 		 penum->fstack.items[penum->fstack.depth].font);
-	if ( penum->can_cache )
+	/* can_cache >= 0 allows us to use cached characters, */
+	/* even if we can't make new cache entries. */
+	if ( penum->can_cache >= 0 )
 	{	/* Loop with cache */
-		if ( pair == 0 )
-			pair = gx_lookup_fm_pair(pfont, pgs);
 		for ( ; ; )
 		{   switch ( (code = (*next_char)(penum, &chr)) )
 		    {
@@ -754,14 +824,24 @@ more:	/* Proceed to next character */
 			pfont = penum->fstack.items[penum->fstack.depth].font;
 			pgs->char_tm_valid = false;
 			show_state_setup(penum);
-			pair = gx_lookup_fm_pair(pfont, pgs);
+			pair = 0;
 			/* falls through */
 		    case 0:				/* plain char */
+			/*
+			 * We don't need to set penum->current_char in the
+			 * normal cases, but it's needed for widthshow,
+			 * kshow, and one strange client, so we may as well
+			 * do it here.
+			 */
+			penum->current_char = chr;
 			glyph = (*penum->encode_char)(penum, pfont, &chr);
+			penum->current_char = chr;
 			if ( glyph == gs_no_glyph )
 			  {	cc = 0;
 				goto no_cache;
 			  }
+			if ( pair == 0 )
+			  pair = gx_lookup_fm_pair(pfont, pgs);
 			cc = gx_lookup_cached_char(pfont, pair, glyph, wmode,
 						   alpha_bits);
 			if ( cc == 0 )
@@ -774,18 +854,20 @@ more:	/* Proceed to next character */
 			      goto no_cache;
 			    if ( pfont->BitmapWidths )
 			      { cc = gx_lookup_xfont_char(pgs, pair, chr,
-					glyph, pfont->procs.glyph_name, wmode);
+					glyph, &pfont->procs.callbacks, wmode);
 				if ( cc == 0 )
 				  goto no_cache;
 			      }
 			    else
-			      { if ( penum->stringwidth_flag != 0 )
+			      { if ( penum->stringwidth_flag != 0 ||
+				     penum->charpath_flag != cpm_show
+				   )
 				  goto no_cache;
 				/* We might have an xfont, but we still */
 				/* want the scalable widths. */
 				cc = gx_lookup_xfont_char(pgs, pair, chr,
-					glyph, pfont->procs.glyph_name, wmode);
-				/* Run the scaler up to the point of */
+					glyph, &pfont->procs.callbacks, wmode);
+				/* Render up to the point of */
 				/* setcharwidth or setcachedevice, */
 				/* just as for stringwidth. */
 				/* This is the only case in which we can */
@@ -794,7 +876,39 @@ more:	/* Proceed to next character */
 			      }
 			  }
 			/* Character is in cache. */
-			if ( !penum->stringwidth_flag )
+			/* We might be doing .charboxpath or stringwidth; */
+			/* check for these now. */
+			if ( penum->charpath_flag != cpm_show )
+			  { /* This is .charboxpath.  Get the bounding box */
+			    /* and append it to a path. */
+			    gx_path box_path;
+			    gs_fixed_point pt;
+			    fixed llx, lly, urx, ury;
+			    code = gx_path_current_point(pgs->path, &pt);
+			    if ( code < 0 )
+			      return code;
+			    llx = fixed_rounded(pt.x - cc->offset.x) +
+			      int2fixed(penum->ftx);
+			    lly = fixed_rounded(pt.y - cc->offset.y) +
+			      int2fixed(penum->fty);
+			    urx = llx + int2fixed(cc->width),
+			    ury = lly + int2fixed(cc->height);
+			    gx_path_init(&box_path, pgs->memory);
+			    code =
+			      gx_path_add_rectangle(&box_path, llx, lly,
+						    urx, ury);
+			    if ( code >= 0 )
+			      code =
+				gx_path_add_char_path(pgs->show_gstate->path,
+						      &box_path,
+						      penum->charpath_flag);
+			    if ( code >= 0 )
+			      code = gx_path_add_point(pgs->path, pt.x, pt.y);
+			    gx_path_release(&box_path);
+			    if ( code < 0 )
+			      return code;
+			  }
+			else if ( !penum->stringwidth_flag )
 			  { code = gx_image_cached_char(penum, cc);
 			    if ( code < 0 )
 			      return code;
@@ -804,18 +918,16 @@ more:	/* Proceed to next character */
 			      }
 			  }
 			if ( penum->slow_show )
-			  { penum->current_char = chr; /* for widthshow */
-			    /* Split up the assignment so that the */
+			  { /* Split up the assignment so that the */
 			    /* Watcom compiler won't reserve esi/edi. */
 			    penum->wxy.x = cc->wxy.x;
 			    penum->wxy.y = cc->wxy.y;
 			    code = show_move(penum);
 			  }
 			else
-				code = show_fast_move(cc->wxy, pgs);
+			  code = show_fast_move(cc->wxy, pgs);
 			if ( code )
 			  { /* Might be kshow, so store the state. */
-			    penum->current_char = chr;
 			    penum->current_glyph = glyph;
 			    return code;
 			  }
@@ -836,7 +948,9 @@ more:	/* Proceed to next character */
 		case 0:
 			;
 		}
+		penum->current_char = chr;
 		glyph = (*penum->encode_char)(penum, pfont, &chr);
+		penum->current_char = chr;
 		cc = 0;
 	}
 no_cache:
@@ -845,9 +959,9 @@ no_cache:
 	 * we only do this if the character is not cached (cc = 0);
 	 * however, we also must do this if we have an xfont but
 	 * are using scalable widths.  In this case, and only this case,
-	 * we get here with cc != 0.
+	 * we get here with cc != 0.  penum->current_char has already
+	 * been set, but not penum->current_glyph.
 	 */
-	penum->current_char = chr;
 	penum->current_glyph = glyph;
 	if ( (code = gs_gsave(pgs)) < 0 )
 	  return code;
@@ -866,9 +980,10 @@ no_cache:
 	   {	gs_fixed_point cpt;
 		gx_path *ppath = pgs->path;
 		if ( (code = gx_path_current_point_inline(ppath, &cpt)) < 0 )
-		  return code;
-		/* We know that char_tm is valid because of */
-		/* show_state_setup. */
+		  goto rret;
+		/* Normally, char_tm is valid because of show_state_setup, */
+		/* but if we're in a cshow, it may not be. */
+		gs_currentcharmatrix(pgs, NULL, true);
 #if 1 /*USE_FPU <= 0*/
 		if ( pgs->ctm.txy_fixed_valid && pgs->char_tm.txy_fixed_valid )
 		  {	fixed tx = pgs->ctm.tx_fixed;
@@ -889,7 +1004,9 @@ no_cache:
 			fpy = fixed2float(cpt.y) + (pgs->ctm.ty - ty);
 #define f_fits_in_fixed(f) f_fits_in_bits(f, fixed_int_bits)
 			if ( !(f_fits_in_fixed(fpx) && f_fits_in_fixed(fpy)) )
-			  return_error(gs_error_limitcheck);
+			  { gs_note_error(code = gs_error_limitcheck);
+			    goto rret;
+			  }
 			cpt.x = float2fixed(fpx);
 			cpt.y = float2fixed(fpy);
 		  }
@@ -899,22 +1016,26 @@ no_cache:
 		code = show_origin_setup(pgs, cpt.x, cpt.y,
 					 penum->charpath_flag);
 		if ( code < 0 )
-		  return code;
+		  goto rret;
 	   }
 	penum->width_status = sws_none;
 	penum->continue_proc = continue_show_update;
 	/* Try using the build procedure in the font. */
 	/* < 0 means error, 0 means success, 1 means failure. */
+	penum->cc = cc;		/* set this now for build procedure */
 	code = (*pfont->procs.build_char)(penum, pgs, pfont, chr, glyph);
 	if ( code < 0 )
-	  return_error(code);
+	  { discard(gs_note_error(code));
+	    goto rret;
+	  }
 	if ( code == 0 )
 	   {	code = show_update(penum);
 		if ( code < 0 )
-		  return code;
+		  goto rret;
+		/* Note that show_update does a grestore.... */
 		code = show_move(penum);
 		if ( code )
-		  return code;
+		  return code;	/* ... so don't go to rret here. */
 		goto more;
 	   }
 	/*
@@ -926,10 +1047,13 @@ no_cache:
 	if ( penum->dev_cache == 0 )
 	  {	code = show_cache_setup(penum);
 		if ( code < 0 )
-		  return code;
+		  goto rret;
 	  }
-	penum->cc = cc;
 	return gs_show_render;
+	/* If we get an error while setting up for BuildChar, */
+	/* we must undo the partial setup. */
+rret:	gs_grestore(pgs);
+	return code;
 }
 
 /* Finish show or stringwidth */
@@ -963,7 +1087,8 @@ gs_show_current_glyph(const gs_show_enum *penum)
 /* Return the width of the just-enumerated character (for cshow). */
 int
 gs_show_current_width(const gs_show_enum *penum, gs_point *ppt)
-{	return gs_idtransform(penum->pgs, fixed2float(penum->wxy.x),
+{	return gs_idtransform(penum->pgs,
+			      fixed2float(penum->wxy.x),
 			      fixed2float(penum->wxy.y), ppt);
 }
 
@@ -981,18 +1106,25 @@ gs_kshow_next_char(const gs_show_enum *penum)
 
 /* ------ Miscellaneous accessors ------ */
 
+/* Return the current font for cshow. */
+gs_font *
+gs_show_current_font(const gs_show_enum *penum)
+{	return (penum->fstack.depth < 0 ? penum->pgs->font :
+		penum->fstack.items[penum->fstack.depth].font);
+}
+
+/* Return the charpath mode. */
+gs_char_path_mode
+gs_show_in_charpath(const gs_show_enum *penum)
+{	return penum->charpath_flag;
+}
+
 /* Return the accumulated width for stringwidth. */
 void
 gs_show_width(const gs_show_enum *penum, gs_point *ppt)
 {	*ppt = penum->width;
 }
 
-/* Return the charpath flag. */
-int
-gs_show_in_charpath(const gs_show_enum *penum)
-{	return penum->charpath_flag;
-}
-
 /* Return true if we only need the width from the rasterizer */
 /* and can short-circuit the full rendering of the character, */
 /* false if we need the actual character bits. */
@@ -1023,16 +1155,18 @@ show_setup(register gs_show_enum *penum, gs_state *pgs, const char *str)
 	penum->wchr = gs_no_char;
 	penum->add = false;
 	penum->do_kern = 0;
-	penum->charpath_flag = 0;
+	penum->charpath_flag = cpm_show;
 	penum->stringwidth_flag = 0;
 	penum->dev_cache = 0;
+	penum->dev_cache2 = 0;
 	penum->dev_null = 0;
 	penum->index = 0;
+	penum->cc = 0;
 	penum->continue_proc = continue_show;
 	code = (*pfont->procs.init_fstack)(penum, pfont);
 	if ( code < 0 )
 	  return code;
-	penum->can_cache = true;	/* show_state_setup may reset */
+	penum->can_cache = 1;		/* show_state_setup may reset */
 	code = show_state_setup(penum);
 	if ( code < 0 )
 	  return code;
@@ -1057,11 +1191,11 @@ show_state_setup(gs_show_enum *penum)
 		const gx_font_stack_item *pfsi =
 		  &penum->fstack.items[penum->fstack.depth];
 		pfont = pfsi->font;
-		gs_matrix_multiply(&pfsi[-1].font->FontMatrix,
-				   &pfont->FontMatrix, &mat);
+		gs_matrix_multiply(&pfont->FontMatrix,
+				   &pfsi[-1].font->FontMatrix, &mat);
 		gs_setcharmatrix(pgs, &mat);
 	}
-	if ( (penum->can_cache
+	if ( (penum->can_cache >= 0
 #if !CACHE_ROTATED_CHARS
 	      &= /* no skewing or non-rectangular rotation */
 	      (is_fzero2(pgs->char_tm.xy, pgs->char_tm.yx) ||
@@ -1113,7 +1247,7 @@ show_set_scale(gs_show_enum *penum)
 {	/* Decide whether to oversample. */
 	/* We have to decide this each time setcachedevice is called. */
 	const gs_state *pgs = penum->pgs;
-	if ( !penum->charpath_flag &&
+	if ( penum->charpath_flag == cpm_show &&
 	     !penum->stringwidth_flag &&
 	     gx_path_is_void_inline(pgs->path) &&
 		/* Oversampling rotated characters doesn't work well. */
@@ -1127,10 +1261,22 @@ show_set_scale(gs_show_enum *penum)
 			pfont->FontBBox.q.y - pfont->FontBBox.p.y,
 			&extent);
 		if ( code >= 0 )
-		  {	penum->log2_suggested_scale.x =
-			  (extent.x != 0 && any_abs(extent.x) < int2fixed(16) ? 2 : 0);
-			penum->log2_suggested_scale.y =
-			  (extent.y != 0 && any_abs(extent.y) < int2fixed(16) ? 2 : 0);
+		  {	int sx =
+			  (extent.x == 0 ? 0 :
+			   any_abs(extent.x) < int2fixed(25) ? 2 :
+			   any_abs(extent.x) < int2fixed(60) ? 1 :
+			   0);
+			int sy =
+			  (extent.y == 0 ? 0 :
+			   any_abs(extent.y) < int2fixed(25) ? 2 :
+			   any_abs(extent.y) < int2fixed(60) ? 1 :
+			   0);
+			/* If we oversample at all, make sure we do it */
+			/* in both X and Y. */
+			if ( sx == 0 && sy != 0 ) sx = 1;
+			else if ( sy == 0 && sx != 0 ) sy = 1;
+			penum->log2_suggested_scale.x = sx;
+			penum->log2_suggested_scale.y = sy;
 			return;
 		  }
 	  }
@@ -1140,19 +1286,32 @@ show_set_scale(gs_show_enum *penum)
 }
 
 /* Set up the cache device and related information. */
+/* Note that we always allocate both cache devices, */
+/* even if we only use one of them. */
 private int
 show_cache_setup(gs_show_enum *penum)
 {	gs_state *pgs = penum->pgs;
 	gs_memory_t *mem = pgs->memory;
 	gx_device_memory *dev =
 	  gs_alloc_struct(mem, gx_device_memory, &st_device_memory,
-			  "show_cache_setup");
-	if ( dev == 0 )
-	  return_error(gs_error_VMerror);
+			  "show_cache_setup(dev_cache)");
+	gx_device_memory *dev2 =
+	  gs_alloc_struct(mem, gx_device_memory, &st_device_memory,
+			  "show_cache_setup(dev_cache2)");
+	if ( dev == 0 || dev2 == 0 )
+	  { gs_free_object(mem, dev2, "show_cache_setup(dev_cache2)");
+	    gs_free_object(mem, dev, "show_cache_setup(dev_cache)");
+	    return_error(gs_error_VMerror);
+	  }
+	/* We only initialize the device for the sake of the GC, */
+	/* (since we have to re-initialize it as either a mem_mono */
+	/* or a mem_abuf device before actually using it) and also */
+	/* to set its memory pointer. */
 	gs_make_mem_mono_device(dev, mem, gs_currentdevice_inline(pgs));
-	dev->cached.white = 1;
-	dev->cached.black = 1;
 	penum->dev_cache = dev;
+	penum->dev_cache2 = dev2;
+	/* Initialize dev2 for the sake of the GC. */
+	*dev2 = *dev;
 	return 0;
 }
 
@@ -1160,8 +1319,9 @@ show_cache_setup(gs_show_enum *penum)
 /* Used before rendering characters, and for moving the origin */
 /* in setcachedevice2 when WMode=1. */
 private int
-show_origin_setup(gs_state *pgs, fixed cpt_x, fixed cpt_y, int charpath_flag)
-{	if ( !charpath_flag )
+show_origin_setup(gs_state *pgs, fixed cpt_x, fixed cpt_y,
+  gs_char_path_mode charpath_flag)
+{	if ( charpath_flag == cpm_show )
 	  { /* Round the translation in the graphics state. */
 	    /* This helps prevent rounding artifacts later. */
 	    cpt_x = fixed_rounded(cpt_x);
