@@ -1,13 +1,24 @@
-/* Copyright (C) 1993, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1993, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* isave.c */
@@ -18,7 +29,7 @@
 #include "gsexit.h"
 #include "gsstruct.h"
 #include "iastate.h"
-#include "iname.h"
+#include "inamedef.h"
 #include "ipacked.h"
 #include "isave.h"
 #include "isstate.h"
@@ -39,6 +50,10 @@ private_st_alloc_save();
 /* see below for details. */
 private const long max_repeated_scan = 100000;
 
+#define print_save(str, space, sav)\
+  if_debug5('u', "[u]%s space %u 0x%lx: cdata = 0x%lx, id = %lu\n",\
+    str, space, (ulong)(sav), (ulong)(sav)->client_data, (ulong)(sav)->id);
+
 /*
  * The logic for saving and restoring the state is complex.
  * Both the changes to individual objects, and the overall state
@@ -286,17 +301,14 @@ alloc_save_state(gs_dual_memory_t *dmem, void *cdata)
 	if ( gsave != 0 )
 	  { gsave->id = sid + 1;
 	    gsave->client_data = 0;
-	    /* Do the right thing about names. */
-	    if ( gsave->name_cnt != max_name_count )
-	      lsave->name_cnt = gsave->name_cnt,
-	      gsave->name_cnt = max_name_count;
+	    print_save("save", gmem->space, gsave);
+	    /* Restore names when we do the local restore. */
+	    lsave->restore_names = gsave->restore_names;
+	    gsave->restore_names = false;
 	  }
-#define print_save(sav)\
-  if_debug3('u', "[u]save 0x%lx: cdata = 0x%lx, id = %lu\n",\
-    (ulong)(sav), (ulong)(sav)->client_data, (ulong)(sav)->id);
 	lsave->id = sid;
 	lsave->client_data = cdata;
-	print_save(lsave);
+	print_save("save", lmem->space, lsave);
 	/* Reset the l_new attribute in all slots.  The only slots that */
 	/* can have the attribute set are the ones on the changes chain, */
 	/* and ones in objects allocated since the last save. */
@@ -311,7 +323,7 @@ alloc_save_state(gs_dual_memory_t *dmem, void *cdata)
 			if ( rsave != 0 )
 			  {	rsave->id = sid;
 				rsave->client_data = cdata;
-				print_save(rsave);
+				print_save("save", lmem->space, rsave);
 				lsave->id = 0;	/* mark as invisible */
 				lsave->client_data = 0;
 				/* Inherit the allocated space count -- */
@@ -320,11 +332,10 @@ alloc_save_state(gs_dual_memory_t *dmem, void *cdata)
 				  lsave->state.allocated +
 				  lsave->state.inherited;
 				lmem->inherited = rsave->state.inherited;
-				print_save(lsave);
+				print_save("save", lmem->space, lsave);
 			  }
 		  }
 	  }
-#undef print_save
 	dmem->save_level++;
 	set_in_save(dmem);
 	/* Notify the file machinery we just did a save. */
@@ -385,9 +396,7 @@ alloc_save_space(gs_ref_memory_t *mem, gs_dual_memory_t *dmem)
 	  }
 	save->state = save_mem;
 	save->dmem = dmem;
-	save->name_cnt =
-	  (name_memory() == (gs_memory_t *)mem ? name_count() :
-	   max_name_count);
+	save->restore_names = (name_memory() == (gs_memory_t *)mem);
 	save->is_current = (dmem->current == mem);
 	mem->saved = save;
 	return save;
@@ -443,6 +452,16 @@ alloc_save_level(const gs_dual_memory_t *dmem)
 {	return dmem->save_level;
 }
 
+/* Return the innermost externally visible save object, i.e., */
+/* the innermost save with a non-zero ID. */
+alloc_save_t *
+alloc_save_current(const gs_dual_memory_t *dmem)
+{	alloc_save_t *save = dmem->space_local->saved;
+	while ( save != 0 && save->id == 0 )
+	  save = save->state.saved;
+	return save;
+}
+
 /* Test whether a reference would be invalidated by a restore. */
 bool
 alloc_is_since_save(const void *vptr, const alloc_save_t *save)
@@ -508,11 +527,26 @@ alloc_is_since_save(const void *vptr, const alloc_save_t *save)
 /* Test whether a name would be invalidated by a restore. */
 bool
 alloc_name_is_since_save(const ref *pnref, const alloc_save_t *save)
-{	return name_is_since_count(pnref, save->name_cnt);
+{	const name *pname;
+	if ( !save->restore_names )
+	  return false;
+	pname = pnref->value.pname;
+	if ( pname->foreign_string )
+	  return false;
+	return alloc_is_since_save(pname->string_bytes, save);
 }
 bool
 alloc_name_index_is_since_save(uint nidx, const alloc_save_t *save)
-{	return name_index_is_since_count(nidx, save->name_cnt);
+{	ref nref;
+	nref.value.pname = name_index_ptr(nidx);
+	return alloc_name_is_since_save(&nref, save);
+}
+
+/* Check whether any names have been created since a given save */
+/* that might be released by the restore. */
+bool
+alloc_any_names_since_save(const alloc_save_t *save)
+{	return save->restore_names;
 }
 
 /* Get the saved state with a given ID. */
@@ -535,18 +569,25 @@ alloc_save_client_data(const alloc_save_t *save)
 {	return save->client_data;
 }
 
-/* Restore the state.  The client is responsible for calling */
-/* alloc_find_save to get the save object, and for ensuring that */
-/* there are no surviving pointers for which alloc_is_since_save is true. */
+/*
+ * Do one step of restoring the state.  The client is responsible for
+ * calling alloc_find_save to get the save object, and for ensuring that
+ * there are no surviving pointers for which alloc_is_since_save is true.
+ * Return true if the argument was the innermost save, in which case
+ * this is the last (or only) step.
+ * Note that "one step" may involve multiple internal steps,
+ * if this is the outermost restore (which requires restoring both local
+ * and global VM) or if we created extra save levels to reduce scanning.
+ */
 private void restore_finalize(P1(gs_ref_memory_t *));
 private void restore_space(P1(gs_ref_memory_t *));
-void
-alloc_restore_state(alloc_save_t *save)
+bool
+alloc_restore_state_step(alloc_save_t *save)
 {	gs_dual_memory_t *dmem = save->dmem;
 	gs_ref_memory_t *mem = dmem->space_local;
 	alloc_save_t *sprev;
 
-	/* Iteratively restore the state. */
+	/* Do one (externally visible) step of restoring the state. */
 	do
 	{	ulong sid;
 		sprev = mem->saved;
@@ -555,7 +596,9 @@ alloc_restore_state(alloc_save_t *save)
 		restore_resources(sprev, mem);	/* release other resources */
 		restore_space(mem);		/* release memory */
 		if ( sid != 0 )
-		  dmem->save_level--;
+		  {	dmem->save_level--;
+			break;
+		  }
 	}
 	while ( sprev != save );
 
@@ -574,6 +617,8 @@ alloc_restore_state(alloc_save_t *save)
 	{	/* Set the l_new attribute in all slots that are now new. */
 		save_set_new(mem, true);
 	}
+
+	return sprev == save;
 }
 /* Restore the memory of one space, by undoing changes and freeing */
 /* memory allocated since the save. */
@@ -582,9 +627,7 @@ restore_space(gs_ref_memory_t *mem)
 {	alloc_save_t *save = mem->saved;
 	alloc_save_t saved;
 
-	if_debug4('u', "[u]restore space %u from 0x%lx, cdata = 0x%lx, id = %lu\n",
-		  mem->space, (ulong)save, (ulong)save->client_data,
-		  (ulong)save->id);
+	print_save("restore", mem->space, save);
 
 	/* Undo changes since the save. */
 	{	register alloc_change_t *cp = mem->changes;
@@ -632,7 +675,7 @@ alloc_restore_all(gs_dual_memory_t *dmem)
 {
 	/* Restore to a state outside any saves. */
 	while ( dmem->save_level != 0 )
-	  alloc_restore_state(dmem->space_local->saved);
+	  discard(alloc_restore_state_step(dmem->space_local->saved));
 
 	/* Finalize memory. */
 	restore_finalize(dmem->space_local);
@@ -646,7 +689,7 @@ alloc_restore_all(gs_dual_memory_t *dmem)
 	/* save and memory objects. */
 	{	alloc_save_t empty_save;
 		empty_save.dmem = dmem;
-		empty_save.name_cnt = max_name_count;	/* don't bother to release */
+		empty_save.restore_names = false;	/* don't bother to release */
 		restore_resources(&empty_save, NULL);
 	}
 
@@ -662,12 +705,17 @@ alloc_restore_all(gs_dual_memory_t *dmem)
 
 }
 
-/* Finalize objects that will be freed by a restore. */
+/*
+ * Finalize objects that will be freed by a restore.
+ * Note that we must temporarily disable the freeing operations
+ * of the allocator while doing this.
+ */
 private void
 restore_finalize(gs_ref_memory_t *mem)
 {	chunk_t *cp;
 
 	alloc_close_chunk(mem);
+	gs_enable_free((gs_memory_t *)mem, false);
 	for ( cp = mem->clast; cp != 0; cp = cp->cprev )
 	  {	SCAN_CHUNK_OBJECTS(cp)
 		  DO_ALL
@@ -681,6 +729,7 @@ restore_finalize(gs_ref_memory_t *mem)
 			  }
 		END_OBJECTS_SCAN
 	  }
+	gs_enable_free((gs_memory_t *)mem, true);
 }
 
 /* Release resources for a restore */
@@ -694,7 +743,8 @@ restore_resources(alloc_save_t *sprev, gs_ref_memory_t *mem)
 	font_restore(sprev);
 
 	/* Adjust the name table. */
-	name_restore(sprev->name_cnt);
+	if ( sprev->restore_names )
+	  name_restore(sprev);
 }
 
 /* Release memory for a restore. */
@@ -722,6 +772,8 @@ alloc_forget_save(alloc_save_t *save)
 	gs_ref_memory_t *mem = dmem->space_local;
 	alloc_save_t *sprev;
 
+	print_save("forget_save", mem->space, save);
+
 	/* Iteratively combine the current level with the previous one. */
 	do
 	{	sprev = mem->saved;
@@ -743,6 +795,7 @@ alloc_forget_save(alloc_save_t *save)
 		  }
 		else
 		  {	forget_changes(mem);
+			save_set_new(mem, false);
 			file_forget_save(sprev);
 			combine_space(mem);	/* combine memory */
 			/* This is the outermost save, which might also */
@@ -750,6 +803,7 @@ alloc_forget_save(alloc_save_t *save)
 			mem = dmem->space_global;
 			if ( mem != dmem->space_local && mem->saved != 0 )
 			  {	forget_changes(mem);
+				save_set_new(mem, false);
 				combine_space(mem);
 			  }
 			set_not_in_save(dmem);
@@ -779,12 +833,19 @@ combine_space(gs_ref_memory_t *mem)
 			  mem->pcc = outer;
 			if ( mem->cfreed.cp == cp )
 			  mem->cfreed.cp = outer;
-			/* "Free" the header of the inner chunk. */
-			{ obj_header_t *hp = (obj_header_t *)cp->chead;
+			/* "Free" the header of the inner chunk, */
+			/* and any immediately preceding gap left by */
+			/* the GC having compacted the outer chunk. */
+			{ obj_header_t *hp = (obj_header_t *)outer->cbot;
 			  hp->o_large = 0;
-			  hp->o_size =
-			    sizeof(chunk_head_t) - sizeof(obj_header_t);
+			  hp->o_size = (char *)(cp->chead + 1)
+                            - (char *)(hp + 1);
 			  hp->o_type = &st_bytes;
+			  /* The following call is probably not safe. */
+#if 0				/* **************** */
+                          gs_free_object((gs_memory_t *)mem,
+                                         hp + 1, "combine_space(header)");
+#endif				/* **************** */
 			}
 			/* Update the outer chunk's allocation pointers. */
 			outer->cbot = cp->cbot;
@@ -823,13 +884,18 @@ combine_space(gs_ref_memory_t *mem)
 	gs_free_object((gs_memory_t *)mem, saved, "combine_space(saved)");
 	alloc_open_chunk(mem);
 }
-/* Free the changes chain. */
+/* Free the changes chain for a level 0 .forgetsave, */
+/* resetting the l_new flag in the changed refs. */
 private void
 forget_changes(gs_ref_memory_t *mem)
 {	register alloc_change_t *chp = mem->changes;
 	alloc_change_t *next;
 	for ( ; chp; chp = next )
-	{	next = chp->next;
+	{	ref_packed *prp = chp->where;
+		if_debug1('U', "[U]forgetting change 0x%lx\n", (ulong)chp);
+		if ( !r_is_packed(prp) )
+		  r_clear_attrs((ref *)prp, l_new);
+		next = chp->next;
 		gs_free_object((gs_memory_t *)mem, chp, "forget_changes");
 	}
 	mem->changes = 0;
