@@ -1,13 +1,24 @@
-/* Copyright (C) 1989, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* zfileio.c */
@@ -75,7 +86,7 @@ zread(register os_ptr op)
 }
 
 /* <file> <int> write - */
-private int
+int
 zwrite(register os_ptr op)
 {	stream *s;
 	byte ch;
@@ -261,6 +272,15 @@ zreadstring_at(register os_ptr op, uint start)
 		return handle_read_status(status, op - 1, &rlen,
 					  zreadstring_continue);
 	}
+	/*
+	 * The most recent Adobe specification says that readstring
+	 * must signal a rangecheck if the string length is zero.
+	 * I can't imagine the motivation for this, but we emulate it.
+	 * It's safe to check it here, rather than earlier, because if
+	 * len is zero, sgets will return 0 immediately with rlen = 0.
+	 */
+	if ( len == 0 )
+	  return_error(e_rangecheck);
 	r_set_size(op, rlen);
 	op[-1] = *op;
 	make_bool(op, (rlen == len ? 1 : 0));
@@ -356,13 +376,15 @@ zreadline(register os_ptr op)
 /* *op is the index within the string, or 0 for an interrupt after a CR. */
 private int
 zreadline_continue(register os_ptr op)
-{	uint start;
+{	uint size = r_size(op - 1);
+	uint start;
 	int code;
+
 	check_type(*op, t_integer);
-	if ( op->value.intval < 0 || op->value.intval > r_size(op - 1) )
+	if ( op->value.intval < 0 || op->value.intval > size )
 	  return_error(e_rangecheck);
 	start = (uint)op->value.intval;
-	code = (start == 0 ? zreadline_at(op - 1, 0, true) :
+	code = (start == 0 ? zreadline_at(op - 1, size, true) :
 		zreadline_at(op - 1, start, false));
 	if ( code >= 0 )
 	  pop(1);
@@ -391,19 +413,14 @@ top:	if ( *pin_eol )
 		 * end-of-line sequence.  If we get an EOF here, stop, but
 		 * don't signal EOF now; wait till the next read.
 		 */
-		int ch;
-		while ( sendrp(s) )
-		  {	int status = s->end_status;
-			if ( status == EOFC )
-			  {	*pin_eol = false;
-				return 0;
-			  }
-			else if ( status < 0 )
-			  return status;
-			s_process_read_buf(s);
+		int ch = spgetcc(s, false);
+		if ( ch == EOFC )
+		  { *pin_eol = false;
+		    return 0;
 		  }
-		ch = sgetc(s);	/* can't fail */
-		if ( ch != LF )
+		else if ( ch < 0 )
+		  return ch;
+		else if ( ch != LF )
 		  sputback(s);
 		*pin_eol = false;
 		return 0;
@@ -437,6 +454,7 @@ top:	if ( *pin_eol )
 		}
 		if ( count >= size )	/* filled the string */
 		{	sputback(s);
+			*pcount = count;
 			return 1;
 		}
 		ptr[count++] = ch;
@@ -577,27 +595,29 @@ zunread(register os_ptr op)
 	return 0;
 }
 
-/* <file> <object> .writecvs - */
-private int zwritecvs_continue(P1(os_ptr));
+/* <file> <object> <==flag> .writecvp - */
+private int zwritecvp_continue(P1(os_ptr));
 private int
-zwritecvs_at(register os_ptr op, uint start)
+zwritecvp_at(register os_ptr op, uint start)
 {	stream *s;
 #define max_cvs 128
 	byte str[max_cvs];
 	ref rstr;
-	byte *pchars = str;
+	const byte *pchars = str;
 	uint len;
 	int code, status;
-	check_write_file(s, op - 1);
-	code = obj_cvs(op, str, max_cvs, &len, &pchars);
+
+	check_write_file(s, op - 2);
+	check_type(*op, t_boolean);
+	code = obj_cvp(op - 1, str, max_cvs, &len, &pchars, op->value.boolval);
 	if ( code < 0 )
 	{	if ( pchars == str )
-			return code;
+		  return code;
 	}
 	if ( start > len )
-		return_error(e_rangecheck);
+	  return_error(e_rangecheck);
 	r_set_size(&rstr, len - start);
-	rstr.value.bytes = pchars + start;
+	rstr.value.const_bytes = pchars + start;
 	status = write_string(&rstr, s);
 	switch ( status )
 	{
@@ -608,26 +628,26 @@ zwritecvs_at(register os_ptr op, uint start)
 	case INTC:
 	case CALLC:
 		len -= r_size(&rstr);
-		return handle_write_status(status, op - 1, &len,
-					   zwritecvs_continue);
+		return handle_write_status(status, op - 2, &len,
+					   zwritecvp_continue);
 	}
-	pop(2);
+	pop(3);
 	return 0;
 #undef max_cvs
 }
 private int
-zwritecvs(os_ptr op)
-{	return zwritecvs_at(op, 0);
+zwritecvp(os_ptr op)
+{	return zwritecvp_at(op, 0);
 }
-/* Continue a .writecvs after a callout. */
+/* Continue a .writecvp after a callout. */
 /* *op is the index within the string. */
 private int
-zwritecvs_continue(os_ptr op)
+zwritecvp_continue(os_ptr op)
 {	int code;
 	check_type(*op, t_integer);
 	if ( op->value.intval != (uint)op->value.intval )
 	  return_error(e_rangecheck);
-	code = zwritecvs_at(op - 1, (uint)op->value.intval);
+	code = zwritecvp_at(op - 1, (uint)op->value.intval);
 	if ( code >= 0 )
 	  pop(1);
 	return code;
@@ -652,7 +672,7 @@ BEGIN_OP_DEFS(zfileio_op_defs) {
 	{"2setfileposition", zsetfileposition},
 	{"2unread", zunread},
 	{"2write", zwrite},
-	{"2.writecvs", zwritecvs},
+	{"3.writecvp", zwritecvp},
 	{"2writehexstring", zwritehexstring},
 	{"2writestring", zwritestring},
 END_OP_DEFS(0) }
@@ -662,7 +682,7 @@ END_OP_DEFS(0) }
 /* Switch a file open for read/write access but currently in write mode */
 /* to read mode. */
 int
-file_switch_to_read(ref *op)
+file_switch_to_read(const ref *op)
 {	stream *s = fptr(op);
 	if ( s->write_id != r_size(op) || s->file == 0 )	/* not valid */
 	  return_error(e_invalidaccess);
@@ -676,7 +696,7 @@ file_switch_to_read(ref *op)
 /* Switch a file open for read/write access but currently in read mode */
 /* to write mode. */
 int
-file_switch_to_write(ref *op)
+file_switch_to_write(const ref *op)
 {	stream *s = fptr(op);
 	if ( s->read_id != r_size(op) || s->file == 0 )	/* not valid */
 	  return_error(e_invalidaccess);
@@ -724,16 +744,13 @@ handle_read_status(int ch, const ref *fop, const uint *pindex,
 		return 1;
 	case INTC:
 	case CALLC:
-	{	ref index;
-		const ref *pstate;
 		if ( pindex )
-		  {	make_int(&index, *pindex);
-			pstate = &index;
+		  { ref index;
+		    make_int(&index, *pindex);
+		    return s_handle_read_exception(ch, fop, &index, 1, cont);
 		  }
 		else
-			pstate = NULL;
-		return s_handle_read_exception(ch, fop, pstate, cont);
-	}
+		  return s_handle_read_exception(ch, fop, NULL, 0, cont);
 	}
 }
 
@@ -750,15 +767,12 @@ handle_write_status(int ch, const ref *fop, const uint *pindex,
 		return_error(e_ioerror);
 	case INTC:
 	case CALLC:
-	{	ref index;
-		const ref *pstate;
 		if ( pindex )
-		  {	make_int(&index, *pindex);
-			pstate = &index;
+		  { ref index;
+		    make_int(&index, *pindex);
+		    return s_handle_write_exception(ch, fop, &index, 1, cont);
 		  }
 		else
-			pstate = NULL;
-		return s_handle_write_exception(ch, fop, pstate, cont);
-	}
+		  return s_handle_write_exception(ch, fop, NULL, 0, cont);
 	}
 }
